# cython: language_level=3
# Copyright (c) 2020-2024 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Sequence, Iterable, Iterator
import cython
import warnings
from .vector cimport (
    Vec3,
    Vec2,
    isclose,
    v3_add,
    v3_mul,
    v3_dist,
    v3_lerp,
    v3_from_angle,
    normalize_rad_angle,
)
from .matrix44 cimport Matrix44
from libc.math cimport ceil, tan, M_PI
from .construct import arc_angle_span_deg

if TYPE_CHECKING:
    from ezdxf.math import UVec
    from ezdxf.math.ellipse import ConstructionEllipse

__all__ = [
    'Bezier4P', 'cubic_bezier_arc_parameters',
    'cubic_bezier_from_arc', 'cubic_bezier_from_ellipse',
]

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL
    const double M_TAU

cdef double DEG2RAD = M_PI / 180.0
cdef double RECURSION_LIMIT = 1000


cdef class Bezier4P:
    cdef:
        FastCubicCurve curve    # pyright: ignore
        readonly Vec3 start_point
        Vec3 cp1
        Vec3 cp2
        readonly Vec3 end_point

    def __cinit__(self, defpoints: Sequence[UVec]):
        if not isinstance(defpoints[0], (Vec2, Vec3)):
            warnings.warn(
                DeprecationWarning, 
                "Bezier4P requires defpoints of type Vec2 or Vec3 in the future",
            )
        if len(defpoints) == 4:
            self.start_point = Vec3(defpoints[0])
            self.cp1 = Vec3(defpoints[1])
            self.cp2 = Vec3(defpoints[2])
            self.end_point = Vec3(defpoints[3])

            self.curve = FastCubicCurve(
                self.start_point,
                self.cp1,
                self.cp2,
                self.end_point
            )
        else:
            raise ValueError("Four control points required.")

    @property
    def control_points(self) -> tuple[Vec3, Vec3, Vec3, Vec3]:
        return self.start_point, self.cp1, self.cp2, self.end_point

    def __reduce__(self):
        return Bezier4P, (self.control_points,)

    def point(self, double t) -> Vec3:
        if 0.0 <= t <= 1.0:
            return self.curve.point(t)
        else:
            raise ValueError("t not in range [0 to 1]")

    def tangent(self, double t) -> Vec3:
        if 0.0 <= t <= 1.0:
            return self.curve.tangent(t)
        else:
            raise ValueError("t not in range [0 to 1]")

    def approximate(self, int segments) -> list[Vec3]:
        cdef double delta_t
        cdef int segment
        cdef list points = [self.start_point]

        if segments < 1:
            raise ValueError(segments)
        delta_t = 1.0 / segments
        for segment in range(1, segments):
            points.append(self.curve.point(delta_t * segment))
        points.append(self.end_point)
        return points

    def flattening(self, double distance, int segments = 4) -> list[Vec3]:
        cdef double dt = 1.0 / segments
        cdef double t0 = 0.0, t1
        cdef _Flattening f = _Flattening(self, distance)
        cdef Vec3 start_point = self.start_point
        cdef Vec3 end_point
        while t0 < 1.0:
            t1 = t0 + dt
            if isclose(t1, 1.0, REL_TOL, ABS_TOL):
                end_point = self.end_point
                t1 = 1.0
            else:
                end_point = self.curve.point(t1)
            f.reset_recursion_check()
            f.flatten(start_point, end_point, t0, t1)
            if f.has_recursion_error():
                raise RecursionError(
                    "Bezier4P flattening error, check for very large coordinates"
                )
            t0 = t1
            start_point = end_point
        
        return f.points

    def approximated_length(self, segments: int = 128) -> float:
        cdef double length = 0.0
        cdef bint start_flag = 0
        cdef Vec3 prev_point, point

        for point in self.approximate(segments):
            if start_flag:
                length += v3_dist(prev_point, point)
            else:
                start_flag = 1
            prev_point = point
        return length

    def reverse(self) -> Bezier4P:
        return Bezier4P((self.end_point, self.cp2, self.cp1, self.start_point))

    def transform(self, Matrix44 m) -> Bezier4P:
        return Bezier4P(tuple(m.transform_vertices(self.control_points)))


cdef class _Flattening:
    cdef FastCubicCurve curve  # pyright: ignore
    cdef double distance
    cdef list points
    cdef int _recursion_level
    cdef int _recursion_error

    def __cinit__(self, Bezier4P curve, double distance):
        self.curve = curve.curve
        self.distance = distance
        self.points = [curve.start_point]
        self._recursion_level = 0
        self._recursion_error = 0

    cdef has_recursion_error(self):
        return self._recursion_error

    cdef reset_recursion_check(self):
        self._recursion_level = 0
        self._recursion_error = 0

    cdef flatten(
        self,
        Vec3 start_point,
        Vec3 end_point,
        double start_t,
        double end_t
    ):
        # Keep in sync with CPython implementation: ezdxf/math/_bezier4p.py
        # Test suite: 630a
        if self._recursion_level > RECURSION_LIMIT:
            self._recursion_error = 1
            return
        self._recursion_level += 1
        cdef double mid_t = (start_t + end_t) * 0.5
        cdef Vec3 mid_point = self.curve.point(mid_t)
        cdef double d = v3_dist(mid_point, v3_lerp(start_point, end_point, 0.5))
        if d < self.distance:
            self.points.append(end_point)
        else:
            self.flatten(start_point, mid_point, start_t, mid_t)
            self.flatten(mid_point, end_point, mid_t, end_t)
        self._recursion_level -= 1

cdef double DEFAULT_TANGENT_FACTOR = 4.0 / 3.0  # 1.333333333333333333
cdef double OPTIMIZED_TANGENT_FACTOR = 1.3324407374108935
cdef double TANGENT_FACTOR = DEFAULT_TANGENT_FACTOR

@cython.cdivision(True)
def cubic_bezier_arc_parameters(
    double start_angle, 
    double end_angle,
    int segments = 1,
) -> Iterator[tuple[Vec3, Vec3, Vec3, Vec3]]:
    if segments < 1:
        raise ValueError('Invalid argument segments (>= 1).')
    cdef double delta_angle = end_angle - start_angle
    cdef int arc_count
    if delta_angle > 0:
        arc_count = <int> ceil(delta_angle / M_PI * 2.0)
        if segments > arc_count:
            arc_count = segments
    else:
        raise ValueError('Delta angle from start- to end angle has to be > 0.')

    cdef double segment_angle = delta_angle / arc_count
    cdef double tangent_length = TANGENT_FACTOR * tan(segment_angle / 4.0)
    cdef double angle = start_angle
    cdef Vec3 start_point, end_point, cp1, cp2
    end_point = v3_from_angle(angle, 1.0)

    for _ in range(arc_count):
        start_point = end_point
        angle += segment_angle
        end_point = v3_from_angle(angle, 1.0)
        cp1 = Vec3()
        cp1.x = start_point.x - start_point.y * tangent_length
        cp1.y = start_point.y + start_point.x * tangent_length
        cp2 = Vec3()
        cp2.x = end_point.x + end_point.y * tangent_length
        cp2.y = end_point.y - end_point.x * tangent_length
        yield start_point, cp1, cp2, end_point

def cubic_bezier_from_arc(
    center = (0, 0), 
    double radius = 1.0, 
    double start_angle = 0.0,
    double end_angle = 360.0, 
    int segments = 1
) -> Iterable[Bezier4P]:
    cdef Vec3 center_ = Vec3(center)
    cdef Vec3 tmp
    cdef list res
    cdef int i
    cdef double angle_span = arc_angle_span_deg(start_angle, end_angle)
    if abs(angle_span) < 1e-9:
        return

    cdef double s = start_angle
    start_angle = (s * DEG2RAD) % M_TAU
    end_angle = (s + angle_span) * DEG2RAD
    while start_angle > end_angle:
        end_angle += M_TAU

    for control_points in cubic_bezier_arc_parameters(start_angle, end_angle, segments):
        res = []
        for i in range(4):
            tmp = <Vec3> control_points[i]
            res.append(v3_add(center_, v3_mul(tmp, radius)))
        yield Bezier4P(res)

def cubic_bezier_from_ellipse(
    ellipse: ConstructionEllipse,
    int segments = 1
) -> Iterator[Bezier4P]:
    cdef double param_span = ellipse.param_span
    if abs(param_span) < 1e-9:
        return

    cdef double start_angle = normalize_rad_angle(ellipse.start_param)
    cdef double end_angle = start_angle + param_span

    while start_angle > end_angle:
        end_angle += M_TAU

    cdef Vec3 center = Vec3(ellipse.center)
    cdef Vec3 x_axis = Vec3(ellipse.major_axis)
    cdef Vec3 y_axis = Vec3(ellipse.minor_axis)
    cdef Vec3 cp
    cdef Vec3 c_res
    cdef list res
    for control_points in cubic_bezier_arc_parameters(start_angle, end_angle, segments):
        res = list()
        for i in range(4):
            cp = <Vec3> control_points[i]
            c_res = v3_add_3(center, v3_mul(x_axis, cp.x), v3_mul(y_axis, cp.y))
            res.append(c_res)
        yield Bezier4P(res)


cdef Vec3 v3_add_3(Vec3 a, Vec3 b, Vec3 c):
    cdef Vec3 result = Vec3()

    result.x = a.x + b.x + c.x
    result.y = a.y + b.y + c.y
    result.z = a.z + b.z + c.z

    return result


cdef class FastCubicCurve:
    cdef:
        double[3] offset
        double[3] p1
        double[3] p2
        double[3] p3

    def __cinit__(self, Vec3 p0, Vec3 p1, Vec3 p2, Vec3 p3):
        cdef:
            double x = p0.x
            double y = p0.y
            double z = p0.z

        self.offset[0] = x
        self.offset[1] = y
        self.offset[2] = z

        # 1st control point (p0) is always (0, 0, 0)
        self.p1[0] = p1.x - x
        self.p1[1] = p1.y - y
        self.p1[2] = p1.z - z

        self.p2[0] = p2.x - x
        self.p2[1] = p2.y - y
        self.p2[2] = p2.z - z

        self.p3[0] = p3.x - x
        self.p3[1] = p3.y - y
        self.p3[2] = p3.z - z
        

    cdef Vec3 point(self, double t):
        # 1st control point (p0) is always (0, 0, 0)
        # => p0 * a is always (0, 0, 0)
        cdef:
            Vec3 result = Vec3()
            double t2 = t * t
            double _1_minus_t = 1.0 - t
            # a = (1 - t) ** 3
            double b = 3.0 * _1_minus_t * _1_minus_t * t
            double c = 3.0 * _1_minus_t * t2
            double d = t2 * t
            
        iadd_mul(result, self.p1, b)
        iadd_mul(result, self.p2, c)
        iadd_mul(result, self.p3, d)

        # add offset at last - it is maybe very large
        result.x += self.offset[0]
        result.y += self.offset[1]
        result.z += self.offset[2]

        return result


    cdef Vec3 tangent(self, double t):
        # tangent vector is independent from offset location!
        cdef:
            Vec3 result = Vec3()
            double t2 = t * t
            # a = -3 * (1 - t) ** 2
            double b = 3.0 * (1.0 - 4.0 * t + 3.0 * t2)
            double c = 3.0 * t * (2.0 - 3.0 * t)
            double d = 3.0 * t2

        iadd_mul(result, self.p1, b)
        iadd_mul(result, self.p2, c)
        iadd_mul(result, self.p3, d)

        return result


cdef void iadd_mul(Vec3 a, double[3] b, double c):
    a.x += b[0] * c
    a.y += b[1] * c
    a.z += b[2] * c




#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import List
import math
import pytest
from ezdxf.math import ConstructionPolyline, Vec3


@pytest.fixture(scope="module")
def poly1() -> ConstructionPolyline:
    return ConstructionPolyline([(0, 0), (1, 0), (1, 1), (0, 1)])


@pytest.fixture(scope="module")
def poly2() -> ConstructionPolyline:
    return ConstructionPolyline([(0, 0), (1, 0), (1, 1), (0, 1)], close=True)


class TestSequenceInterface:
    def test_empty_polyline(self):
        cp = ConstructionPolyline([])
        assert len(cp) == 0

    def test_len_open_polyline(self, poly1):
        assert len(poly1) == 4

    def test_len_closed_polyline(self, poly2):
        assert len(poly2) == 5

    def test_get_single_vertex(self, poly1):
        assert poly1[1] == (1, 0)

    def test_get_last_vertex(self, poly1):
        assert poly1[-1] == (0, 1)

    def test_get_vertex_slice(self, poly1):
        assert list(poly1[1:3]) == [(1, 0), (1, 1)]

    def test_vertex_slice_has_same_type(self, poly1):
        assert isinstance(poly1[1:3], ConstructionPolyline)

    def test_is_immutable(self, poly1):
        with pytest.raises(TypeError):
            poly1[1] = (1, 0)
        with pytest.raises(TypeError):
            del poly1[1]


class TestExtendedData:
    """The data() method returns the distance from the start vertex along the
    polyline, the distance from the previous vertex and the vertex itself for
    a give index.

    """

    def test_empty_polyline_raises_Value_error(self):
        with pytest.raises(ValueError):
            ConstructionPolyline([]).data(0)

    def test_data_0(self, poly1):
        assert poly1.data(0) == (0, 0, (0, 0))

    def test_data_1(self, poly1):
        assert poly1.data(1) == (1, 1, (1, 0))

    def test_data_2(self, poly1):
        assert poly1.data(2) == (2, 1, (1, 1))

    def test_data_3(self, poly1):
        assert poly1.data(3) == (3, 1, (0, 1))

    def test_data_for_negative_index(self, poly1):
        assert poly1.data(-1) == (3, 1, (0, 1))
        assert poly1.data(-2) == (2, 1, (1, 1))

    def test_vertices_at_the_same_location(self):
        cp = ConstructionPolyline([(0, 0), (0, 0), (0, 0)])
        assert cp.data(0) == (0, 0, (0, 0))
        assert cp.data(1) == (0, 0, (0, 0))
        assert cp.data(2) == (0, 0, (0, 0))

    def test_raises_index_error(self, poly1):
        with pytest.raises(IndexError):
            poly1.data(4)


class TestLength:
    def test_empty_polyline(self):
        assert ConstructionPolyline([]).length == 0.0

    def test_open_polyline(self, poly1):
        assert poly1.length == 3.0

    def test_closed_polyline(self, poly2):
        assert poly2.length == 4.0

    def test_vertices_at_the_same_location(self):
        cp = ConstructionPolyline([(0, 0), (0, 0), (0, 0)])
        assert cp.length == 0.0


class TestIsClosed:
    def test_empty_polyline_is_not_closed(self):
        assert ConstructionPolyline([]).is_closed is False

    def test_polyline_with_too_few_vertices_is_not_closed(self):
        assert ConstructionPolyline([(0, 0)]).is_closed is False
        assert ConstructionPolyline([(0, 0), (0, 0)]).is_closed is False

    def test_poly1_is_not_closed(self, poly1):
        assert poly1.is_closed is False

    def test_poly2_is_closed(self, poly2):
        assert poly2.is_closed is True


class TestIndexAt:
    @pytest.mark.parametrize("d", [-1, 0, 1])
    def test_empty_polyline(self, d):
        assert ConstructionPolyline([]).index_at(d) == 0

    @pytest.mark.parametrize("d", [-1, 0, 1])
    def test_polyline_with_one_vertex(self, d):
        assert ConstructionPolyline([(1, 1)]).index_at(d) == 0

    def test_short_polyline(self):
        poly = ConstructionPolyline([(0, 0), (1, 0)])
        assert poly.index_at(0.0) == 0
        assert poly.index_at(0.5) == 1
        assert poly.index_at(1) == 1

    def test_long_polyline(self):
        poly = ConstructionPolyline([(0, 0), (1, 0), (2, 0), (3, 0)])
        assert poly.index_at(0.0) == 0
        assert poly.index_at(0.999) == 1
        assert poly.index_at(1.0) == 1
        assert poly.index_at(1.001) == 2
        assert poly.index_at(1.999) == 2
        assert poly.index_at(2.0) == 2
        assert poly.index_at(2.001) == 3
        assert poly.index_at(2.999) == 3
        assert poly.index_at(3.0) == 3
        assert poly.index_at(3.001) == 3


class TestVertexAt:
    def test_empty_polyline_raises_error(self):
        with pytest.raises(ValueError):
            ConstructionPolyline([]).vertex_at(0.0)

    def test_too_few_vertices_raises_error(self):
        with pytest.raises(ValueError):
            ConstructionPolyline([(0, 0)]).vertex_at(0.0)

    def test_out_of_range_raises_error(self, poly1):
        with pytest.raises(ValueError):
            poly1.vertex_at(-1)
        with pytest.raises(ValueError):
            poly1.vertex_at(poly1.length + 0.01)

    def test_interpolation_at_vertex_location(self, poly1):
        assert poly1.vertex_at(0.0) == (0, 0)
        assert poly1.vertex_at(1.0) == (1, 0)
        assert poly1.vertex_at(2.0) == (1, 1)
        assert poly1.vertex_at(3.0) == (0, 1)

    def test_interpolate_last_vertex_of_closed_polyline(self, poly2):
        assert poly2.vertex_at(0.0) == (0, 0)
        assert poly2.vertex_at(poly2.length) == (0, 0)

    def test_interpolate_at_first_edge(self, poly1):
        """Interpolate edge  (0, 0) -> (1, 0)"""
        assert poly1.vertex_at(0.1).isclose((0.1, 0))
        assert poly1.vertex_at(0.5).isclose((0.5, 0))
        assert poly1.vertex_at(0.9).isclose((0.9, 0))

    def test_interpolate_at_second_edge(self, poly1):
        """Interpolate edge  (1, 0) -> (1, 1)"""
        assert poly1.vertex_at(1.1).isclose((1.0, 0.1))
        assert poly1.vertex_at(1.5).isclose((1.0, 0.5))
        assert poly1.vertex_at(1.9).isclose((1.0, 0.9))

    def test_interpolation_for_coincident_vertices_in_front(self):
        cp = ConstructionPolyline([(0, 0), (0, 0), (0, 0), (1, 0)])
        assert cp.vertex_at(0.0).isclose((0.0, 0.0))
        assert cp.vertex_at(0.1).isclose((0.1, 0.0))
        assert cp.vertex_at(0.5).isclose((0.5, 0.0))
        assert cp.vertex_at(0.9).isclose((0.9, 0.0))
        assert cp.vertex_at(1.0).isclose((1.0, 0.0))

    def test_interpolation_for_coincident_vertices_after(self):
        cp = ConstructionPolyline([(0, 0), (1, 0), (1, 0), (1, 0)])
        assert cp.vertex_at(0.0).isclose((0.0, 0.0))
        assert cp.vertex_at(0.1).isclose((0.1, 0.0))
        assert cp.vertex_at(0.5).isclose((0.5, 0.0))
        assert cp.vertex_at(0.9).isclose((0.9, 0.0))
        assert cp.vertex_at(1.0).isclose((1.0, 0.0))

    def test_interpolation_for_coincident_vertices_between(self):
        cp = ConstructionPolyline(
            [(0, 0), (0.5, 0), (0.5, 0), (0.5, 0), (1, 0)]
        )
        assert cp.vertex_at(0.0).isclose((0.0, 0.0))
        assert cp.vertex_at(0.1).isclose((0.1, 0.0))
        assert cp.vertex_at(0.5).isclose((0.5, 0.0))
        assert cp.vertex_at(0.9).isclose((0.9, 0.0))
        assert cp.vertex_at(1.0).isclose((1.0, 0.0))


class TestDivide:
    @pytest.mark.parametrize("count", [-1, 0, 1])
    def test_raises_error_invalid_count(self, count, poly1):
        with pytest.raises(ValueError):
            list(poly1.divide(count))

    def test_divide_by_3(self, poly1):
        vertices: List[Vec3] = list(poly1.divide(3))
        assert vertices[0].isclose(poly1[0])
        assert vertices[1].isclose((1, 0.5))
        assert vertices[2].isclose(poly1[-1])

    def test_divide_by_length(self, poly1):
        vertices: List[Vec3] = list(poly1.divide_by_length(0.7))
        assert len(vertices) == 5
        assert vertices[-1].isclose(poly1.vertex_at(2.8))

    def test_divide_by_length_force_last_vertex(self, poly1):
        vertices: List[Vec3] = list(
            poly1.divide_by_length(0.7, force_last=True)
        )
        assert len(vertices) == 6
        assert vertices[-1].isclose(poly1[-1])


class TestApproximationAccuracy:
    def test_unit_circle(self):
        from ezdxf.entities import Circle

        # create unit circle
        circle: Circle = Circle.new(
            dxfattribs={"center": (0, 0), "radius": 1.0}
        )
        cp = ConstructionPolyline(circle.flattening(0.01))
        assert cp.is_closed is True
        assert len(cp) > 20
        assert abs(cp.length - math.tau) < 0.02

    def test_unit_circle_by_path(self):
        from ezdxf import path

        p = path.unit_circle()
        cp = ConstructionPolyline(p.flattening(0.01, segments=16))
        assert cp.is_closed is True
        assert len(cp) > 60
        assert abs(cp.length - math.tau) < 0.002


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.lldxf.tagwriter import TagCollector, Tags

TEXTSTYLE_NAME = "TextStyle"
DIMSTYLE_NAME = "DimensionStyle"
FONT_NAME = "any_font.shx"


@pytest.fixture
def doc():
    doc = ezdxf.new("R2010", setup=False)
    doc.styles.new(TEXTSTYLE_NAME, dxfattribs={"font": FONT_NAME})
    doc.dimstyles.new(DIMSTYLE_NAME, dxfattribs={"dimtxsty": TEXTSTYLE_NAME})
    return doc


def test_export_dimtxsty(doc):
    dimstyle = doc.dimstyles.get(DIMSTYLE_NAME)
    style = doc.styles.get(TEXTSTYLE_NAME)
    t = TagCollector()
    dimstyle.export_dxf(t)
    tags = Tags(t.tags)
    dimtxsty_handle = tags.get_first_value(340)
    assert style.dxf.handle == dimtxsty_handle


def test_reload_dimtxsty(doc, tmpdir):
    filename = tmpdir.join("dim_text_style.dxf")
    doc.saveas(filename)
    # reload file
    doc2 = ezdxf.readfile(filename)

    style = doc2.styles.get(TEXTSTYLE_NAME)
    assert style.dxf.font == FONT_NAME

    dimstyle = doc2.dimstyles.get(DIMSTYLE_NAME)
    assert dimstyle.dxf.dimtxsty == TEXTSTYLE_NAME


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License

from ezdxf.lldxf.tags import text_to_multi_tags, multi_tags_to_text

TEST_TEXT = """Lorem ipsum dolor sit amet, consetetur sadipscing elitr, 
sed diam nonumy eirmod tempor invidunt ut labore et dolore 
magna aliquyam erat, sed diam voluptua. At vero eos et accusam 
et justo duo dolores et ea rebum. Stet clita kasd 
gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. 
Lorem ipsum dolor sit amet, consetetur sadipscing 
elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna 
aliquyam erat, sed diam voluptua. At vero 
eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea 
takimata sanctus est Lorem ipsum dolor sit amet.
"""


def test_text_to_multi_tags():
    tags = text_to_multi_tags(TEST_TEXT, code=303, size=40)
    assert len(tags) == 16
    assert tags[0].code == 303
    assert len(tags[0].value) == 40

    text = multi_tags_to_text(tags)
    assert text == TEST_TEXT



# Copyright (c) 2022-2024, Manfred Moitzi
# License: MIT License
# Public API module (interface)
"""
The main goals of this ACIS support library is:

    1. load and parse simple and known ACIS data structures
    2. create and export simple and known ACIS data structures

It is NOT a goal to edit and export arbitrary existing ACIS structures.

    Don't even try it!

This modules do not implement an ACIS kernel!!!
So tasks beyond stitching some flat polygonal faces to a polyhedron or creating
simple curves is not possible.

To all beginners: GO AWAY!

"""
from .const import (
    AcisException,
    ParsingError,
    InvalidLinkStructure,
    ExportError,
)
from .mesh import mesh_from_body, body_from_mesh, vertices_from_body
from .entities import load, export_sat, export_sab, Body
from .dbg import AcisDebugger, dump_sab_as_text
from .dxf import export_dxf, load_dxf
from .cache import AcisCache



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import re
from ezdxf.tools import guid

HEX4 = "[0-9A-F]{4}"
HEX8 = "[0-9A-F]{8}"
HEX12 = "[0-9A-F]{12}"
# "{XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX}"
AUTOCAD_PATTERN = f"{{{HEX8}-{HEX4}-{HEX4}-{HEX4}-{HEX12}}}"


def test_guid_string_matches_autocad_pattern():
    assert re.fullmatch(AUTOCAD_PATTERN, guid()) is not None


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable
from typing_extensions import Self
import logging

from ezdxf.lldxf import const, validator
from ezdxf.lldxf.attributes import (
    DXFAttributes,
    DefSubclass,
    DXFAttr,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.math import UVec, Vec2, Matrix44, Z_AXIS, NULLVEC
from ezdxf.entities import factory
from .dxfentity import SubclassProcessor, base_class
from .dxfobj import DXFObject
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.lldxf.tags import Tags
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["SpatialFilter"]

# The HEADER variable $XCLIPFRAME determines if the clipping path polygon is displayed 
# and plotted:
#   0 - not displayed, not plotted
#   1 - displayed, not plotted
#   2 - displayed and plotted

logger = logging.getLogger("ezdxf")
AcDbFilter = "AcDbFilter"
AcDbSpatialFilter = "AcDbSpatialFilter"

acdb_filter = DefSubclass(AcDbFilter, {})
acdb_spatial_filter = DefSubclass(
    AcDbSpatialFilter,
    {
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        "origin": DXFAttr(11, xtype=XType.point3d, default=NULLVEC),
        "is_clipping_enabled": DXFAttr(
            71, default=1, validator=validator.is_integer_bool, fixer=RETURN_DEFAULT
        ),
        "has_front_clipping_plane": DXFAttr(
            72, default=0, validator=validator.is_integer_bool, fixer=RETURN_DEFAULT
        ),
        "front_clipping_plane_distance": DXFAttr(40, default=0.0),
        "has_back_clipping_plane": DXFAttr(
            73, default=0, validator=validator.is_integer_bool, fixer=RETURN_DEFAULT
        ),
        "back_clipping_plane_distance": DXFAttr(41, default=0.0),
    },
)
acdb_spatial_filter_group_codes = group_code_mapping(acdb_spatial_filter)


@factory.register_entity
class SpatialFilter(DXFObject):
    DXFTYPE = "SPATIAL_FILTER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_filter, acdb_spatial_filter)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self) -> None:
        super().__init__()
        # clipping path vertices in OCS coordinates
        self._boundary_vertices: tuple[Vec2, ...] = tuple()

        # This matrix is the inverse of the original block reference (insert entity)
        # transformation.  The original block reference transformation is the one that
        # is applied to all entities in the block when the block reference is regenerated.
        self._inverse_insert_matrix = Matrix44()

        # This matrix transforms points into the coordinate system of the clip boundary.
        self._transform_matrix = Matrix44()

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        assert isinstance(entity, SpatialFilter)
        # immutable data
        entity._boundary_vertices = self._boundary_vertices
        entity._inverse_insert_matrix = self._inverse_insert_matrix
        entity._transform_matrix = self._transform_matrix

    @property
    def boundary_vertices(self) -> tuple[Vec2, ...]:
        """Returns the clipping path vertices in OCS coordinates."""
        return self._boundary_vertices

    def set_boundary_vertices(self, vertices: Iterable[UVec]) -> None:
        """Set the clipping path vertices in OCS coordinates."""
        self._boundary_vertices = tuple(Vec2(v) for v in vertices)
        if len(self._boundary_vertices) < 2:
            raise const.DXFValueError("2 or more vertices required")

    @property
    def inverse_insert_matrix(self) -> Matrix44:
        """Returns the inverse insert matrix.

        This matrix is the inverse of the original block reference (insert entity)
        transformation.  The original block reference transformation is the one that
        is applied to all entities in the block when the block reference is regenerated.
        """
        return self._inverse_insert_matrix.copy()

    def set_inverse_insert_matrix(self, m: Matrix44) -> None:
        self._inverse_insert_matrix = m.copy()

    @property
    def transform_matrix(self) -> Matrix44:
        """Returns the transform matrix.

        This matrix transforms points into the coordinate system of the clip boundary.
        """
        return self._transform_matrix.copy()

    def set_transform_matrix(self, m: Matrix44) -> None:
        self._transform_matrix = m.copy()

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.find_subclass(AcDbSpatialFilter)
            if tags:
                try:
                    self._load_boundary_data(tags)
                except IndexError:
                    logger.warning(
                        f"Not enough matrix values in SPATIAL_FILTER(#{processor.handle})"
                    )
                processor.fast_load_dxfattribs(
                    dxf, acdb_spatial_filter_group_codes, subclass=tags
                )
            else:
                logger.warning(
                    f"Required subclass 'AcDbSpatialFilter' in object "
                    f"SPATIAL_FILTER(#{processor.handle}) not present"
                )
        return dxf

    def _load_boundary_data(self, tags: Tags) -> None:
        def to_matrix(v: list[float]) -> Matrix44:
            # raises IndexError if not enough matrix values exist
            return Matrix44(
                # fmt: off
                [
                    v[0], v[4], v[8], 0.0,
                    v[1], v[5], v[9], 0.0,
                    v[2], v[6], v[10], 0.0,
                    v[3], v[7], v[11], 1.0,
                ]
                # fmt: on
            )

        self._boundary_vertices = tuple(Vec2(tag.value) for tag in tags.find_all(10))
        matrix_values = [float(tag.value) for tag in tags.find_all(40)]
        # use only last 24 values
        self._inverse_insert_matrix = to_matrix(matrix_values[-24:-12])
        self._transform_matrix = to_matrix(matrix_values[-12:])

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""

        def write_matrix(m: Matrix44) -> None:
            for index in range(3):
                for value in m.get_col(index):
                    tagwriter.write_tag2(40, value)

        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, AcDbFilter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, AcDbSpatialFilter)
        tagwriter.write_tag2(70, len(self._boundary_vertices))
        for vertex in self._boundary_vertices:
            tagwriter.write_vertex(10, vertex)
        self.dxf.export_dxf_attribs(
            tagwriter, ["extrusion", "origin", "is_clipping_enabled"]
        )
        has_front_clipping = self.dxf.has_front_clipping_plane
        tagwriter.write_tag2(72, has_front_clipping)
        if has_front_clipping:  
            # AutoCAD does no accept tag 40, if front clipping is disabled
            tagwriter.write_tag2(40, self.dxf.front_clipping_plane_distance)
        has_back_clipping = self.dxf.has_back_clipping_plane
        tagwriter.write_tag2(73, has_back_clipping)
        if has_back_clipping:  
            # AutoCAD does no accept tag 41, if back clipping is disabled
            tagwriter.write_tag2(41, self.dxf.back_clipping_plane_distance)

        write_matrix(self._inverse_insert_matrix)
        write_matrix(self._transform_matrix)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import ezdxf
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.entities import factory
from ezdxf.entities import ACADProxyEntity
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Matrix44
from ezdxf.protocols import SupportsVirtualEntities, virtual_entities
from ezdxf.entities.copy import CopyNotSupported

@pytest.fixture(scope="module")
def proxy() -> ACADProxyEntity:
    xtags = ExtendedTags.from_text(AEC_DOOR)
    return cast(ACADProxyEntity, factory.load(xtags, ezdxf.new("R2018")))


def test_load_acad_proxy_entity(proxy):
    assert proxy.dxftype() == "ACAD_PROXY_ENTITY"
    assert proxy.proxy_graphic is not None, "expected proxy graphic data"


def test_export_exact_original_data(proxy):
    expected = basic_tags_from_text(AEC_DOOR)
    collector = TagCollector(dxfversion=ezdxf.const.DXF2013, optional=True)
    proxy.export_dxf(collector)
    assert collector.tags == expected, "expected exact same DXF tags"


EXPECTED_VIRTUAL_ENTITY_TYPES = [
    "POLYLINE",
    "POLYLINE",
    "ARC",
    "ARC",
    "POLYLINE",
    "POLYLINE",
]


def test_virtual_entities_support(proxy):
    types = [e.dxftype() for e in proxy.virtual_entities()]
    assert types == EXPECTED_VIRTUAL_ENTITY_TYPES


def test_virtual_sub_entities_source_tracking(proxy):
    result = set(e.source_of_copy for e in proxy.virtual_entities())
    assert len(result) == 1, "only one source of copy expected"
    assert proxy in result, "proxy should be the source of copy"


def test_virtual_entities_if_no_proxy_graphic_exists(proxy):
    data = proxy.proxy_graphic
    proxy.proxy_graphic = None
    assert len(list(proxy.virtual_entities())) == 0
    proxy.proxy_graphic = data


def test_supports_virtual_entities_protocol(proxy):
    assert isinstance(proxy, SupportsVirtualEntities)
    types = [e.dxftype() for e in virtual_entities(proxy)]
    assert types == EXPECTED_VIRTUAL_ENTITY_TYPES


def test_proxy_entity_is_not_transformable(proxy):
    # Transformation is only for the virtual entities possible!
    with pytest.raises(NotImplementedError):
        proxy.transform(Matrix44.translate(1, 1, 1))


def test_proxy_graphic_is_not_copyable(proxy):
    # Has to be copied as virtual entities!
    with pytest.raises(CopyNotSupported):
        proxy.copy()


AEC_DOOR = """  0
ACAD_PROXY_ENTITY
  5
59D
330
1F
100
AcDbEntity
  8
fire door - 60 mins
100
AcDbProxyEntity
 90
498
 91
518
 71
33
 97
155
 70
0
160
728
310
D80200000D0000000C00000010000000050000000C0000000E000000000100000C00000012000000FF7F0000100000002200000000000000000000000C00000010000000050000000C00000012000000000000000C0000000E000000010000006C00000007000000040000008705D40EB308C2404F852AD01EBFC340000000
310
0000000000188DCA03B308C240BF872AD01EA6C34000000000000000003988CA03B33AC240E19617BA1EA6C3400000000000000000A800D40EB33AC240719417BA1EBFC34000000000000000006C000000070000000400000043041686B408C2407A322AD01E11C7400000000000000000B27C1F91B408C2400A302AD01E2A
310
C7400000000000000000D3771F91B43AC2402C3F17BA1E2AC740000000000000000064FF1586B43AC2409C4117BA1E11C74000000000000000006400000004000000A800D40EB33AC240719417BA1EBFC34000000000000000000000000000908A4000000000000000000000000000000000010000000000F0BFD0A43C4F34
310
41DC3EADE1FCFFFFFFEF3F0000000000000000182D4454FB21F93F00000000640000000400000064FF1586B43AC2409C4117BA1E11C74000000000000000000000000000908A4000000000000000000000000000000000010000000000F03FD0A43C4F3441DCBEADE1FCFFFFFFEFBF0000000000000000182D4454FB21F93F
310
000000006C0000000700000004000000A800D40EB33AC240719417BA1EBFC3400000000000000000A36A8A18B33AC2404D9217BA1ED5C340000000000000000038418A18B3E3C340EE9276FE1DD5C34000000000000000003DD7D30EB3E3C340139576FE1DBFC34000000000000000006C000000070000000400000064FF15
310
86B43AC2409C4117BA1E11C740000000000000000069955F7CB43AC240C14317BA1EFBC6400000000000000000FE6B5F7CB4E3C340634476FE1DFBC6400000000000000000F9D51586B4E3C3403E4276FE1D11C7400000000000000000
162
626
311
9267B00330036003200330045003400450032002D0043003200360044002D0034004500300041002D0041004400370042002D003500340033004500420032003400360046003000450035007D00A80
161
1103
310
5430B80950542012A00C46E501D984612017F0E55A03D4D86811A148789E6883B87C56F0FE7FFFFFF79FC56F0FE7FFFFFF79F9A148789E6883B97D52A8000000000004138800000000000204FA000000000001810480A8000000000000092800000000000603920202A8696AA0A8205B88A82A0A82059E206A82A0A8400000
310
000002047A04A82A0A82A0
360
5A1
340
457
 94
0
1001
ACAD
1000
AcRTDxfName
1000
AEC_DOOR
1000
AcRTDxfName
1000
AEC_DOOR
1000
AcRTDxfName
1000
AEC_DOOR
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.dxfgfx import add_entity, replace_entity
from ezdxf.entities import Point


@pytest.fixture(scope="module")
def msp():
    return ezdxf.new().modelspace()


@pytest.fixture(scope="module")
def db(msp):
    return msp.entitydb


def test_add_entity(msp, db):
    point = msp.add_point((0, 0))
    new_point = Point.new(dxfattribs={"location": (3, 3)})
    add_entity(new_point, msp)
    assert point in msp
    assert point.dxf.handle in db
    assert new_point in msp
    assert new_point.dxf.handle in db
    assert point.dxf.handle != new_point.dxf.handle


def test_replace_entity(msp, db):
    point = msp.add_point((0, 0))
    handle = point.dxf.handle

    new_point = Point.new(dxfattribs={"location": (3, 3)})
    replace_entity(point, new_point, msp)
    assert point.is_alive is False
    assert new_point in msp
    assert new_point.dxf.handle in db
    assert new_point.dxf.handle == handle


def test_replace_entity_without_layout(msp, db):
    point = Point.new(dxfattribs={"location": (3, 3)})
    db.add(point)
    handle = point.dxf.handle

    assert point not in msp
    assert point.dxf.handle in db

    new_point = Point.new(dxfattribs={"location": (3, 3)})
    replace_entity(point, new_point, msp)
    assert point.is_alive is False
    assert new_point not in msp
    assert new_point.dxf.handle in db
    assert new_point.dxf.handle == handle


def test_convert_circle_to_ellipse(msp, db):
    circle = msp.add_circle(center=(3, 3), radius=2)
    ellipse = circle.to_ellipse(replace=False)
    assert circle.dxf.handle in db
    assert ellipse.dxftype() == "ELLIPSE"
    assert ellipse.dxf.handle in db
    assert circle in msp
    assert ellipse in msp


def test_replace_circle_by_ellipse(msp, db):
    circle = msp.add_circle(center=(3, 3), radius=2)
    circle_handle = circle.dxf.handle
    ellipse = circle.to_ellipse(replace=True)
    assert circle.is_alive is False
    assert ellipse.dxftype() == "ELLIPSE"
    assert ellipse.dxf.handle in db
    assert ellipse.dxf.handle == circle_handle
    assert ellipse in msp


def test_convert_circle_to_spline(msp, db):
    circle = msp.add_circle(center=(3, 3), radius=2)
    spline = circle.to_spline(replace=False)
    assert circle.dxf.handle in db
    assert spline.dxftype() == "SPLINE"
    assert spline.dxf.handle in db
    assert circle in msp
    assert spline in msp


def test_replace_circle_by_spline(msp, db):
    circle = msp.add_circle(center=(3, 3), radius=2)
    circle_handle = circle.dxf.handle
    spline = circle.to_spline(replace=True)
    assert circle.is_alive is False
    assert spline.dxftype() == "SPLINE"
    assert spline.dxf.handle in db
    assert spline.dxf.handle == circle_handle
    assert spline in msp


def test_convert_ellipse_to_spline(msp, db):
    ellipse = msp.add_ellipse(center=(3, 3), major_axis=(2, 0), ratio=0.5)
    spline = ellipse.to_spline(replace=False)
    assert ellipse.dxf.handle in db
    assert spline.dxftype() == "SPLINE"
    assert spline.dxf.handle in db
    assert ellipse in msp
    assert spline in msp


def test_replace_ellipse_by_spline(msp, db):
    ellipse = msp.add_ellipse(center=(3, 3), major_axis=(2, 0), ratio=0.5)
    ellipse_handle = ellipse.dxf.handle
    spline = ellipse.to_spline(replace=True)
    assert ellipse.is_alive is False
    assert spline.dxftype() == "SPLINE"
    assert spline.dxf.handle in db
    assert spline.dxf.handle == ellipse_handle
    assert spline in msp


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
# Test only basic features of Cython implementation,
# Full testing and compatibility check with Python implementation
# is located in test suite 632.

import pytest

bezier = pytest.importorskip("ezdxf.acc.bezier3p")
Bezier3P = bezier.Bezier3P
from ezdxf.acc.vector import Vec3
from ezdxf.acc.matrix44 import Matrix44

# check functions:

POINTS = Vec3.list([(0.0, 0.0), (5.0, 5.0), (10.0, 0.0)])


def test_default_constructor():
    c = Bezier3P(POINTS)
    assert c.control_points == Vec3.tuple(POINTS)


@pytest.fixture
def curve():
    return Bezier3P(POINTS)


def test_point(curve):
    assert curve.point(0.5) == (5.0, 2.5)


def test_tangent(curve):
    assert curve.tangent(0.5) == (10.0, 0)


def test_approximate(curve):
    points = curve.approximate(10)
    assert len(points) == 11


def test_flattening(curve):
    points = curve.flattening(0.01)
    assert len(points) == 17


def test_approximated_length(curve):
    assert curve.approximated_length(32) == pytest.approx(11.47678475868991)


def test_reverse(curve):
    r = curve.reverse()
    assert r.control_points == Vec3.tuple(reversed(POINTS))


def test_transform(curve):
    m = Matrix44.translate(1, 1, 0)
    r = curve.transform(m)
    assert r.control_points == Vec3.tuple([(1, 1), (6, 6), (11, 1)])


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    Iterator,
    Sequence,
    Callable,
    Optional,
    NamedTuple,
)
from typing_extensions import Self, TypeAlias
import copy
import abc

from ezdxf.math import BoundingBox2d, Matrix44, Vec2, UVec
from ezdxf.npshapes import NumpyPath2d, NumpyPoints2d, EmptyShapeError
from ezdxf.tools import take2
from ezdxf.tools.clipping_portal import ClippingRect

from .backend import BackendInterface, ImageData
from .config import Configuration
from .properties import BackendProperties
from .type_hints import Color


class DataRecord(abc.ABC):
    def __init__(self) -> None:
        self.property_hash: int = 0
        self.handle: str = ""

    @abc.abstractmethod
    def bbox(self) -> BoundingBox2d:
        ...

    @abc.abstractmethod
    def transform_inplace(self, m: Matrix44) -> None:
        ...


class PointsRecord(DataRecord):
    # n=1 point; n=2 line; n>2 filled polygon
    def __init__(self, points: NumpyPoints2d) -> None:
        super().__init__()
        self.points: NumpyPoints2d = points

    def bbox(self) -> BoundingBox2d:
        try:
            return self.points.bbox()
        except EmptyShapeError:
            pass
        return BoundingBox2d()

    def transform_inplace(self, m: Matrix44) -> None:
        self.points.transform_inplace(m)


class SolidLinesRecord(DataRecord):
    def __init__(self, lines: NumpyPoints2d) -> None:
        super().__init__()
        self.lines: NumpyPoints2d = lines

    def bbox(self) -> BoundingBox2d:
        try:
            return self.lines.bbox()
        except EmptyShapeError:
            pass
        return BoundingBox2d()

    def transform_inplace(self, m: Matrix44) -> None:
        self.lines.transform_inplace(m)


class PathRecord(DataRecord):
    def __init__(self, path: NumpyPath2d) -> None:
        super().__init__()
        self.path: NumpyPath2d = path

    def bbox(self) -> BoundingBox2d:
        try:
            return self.path.bbox()
        except EmptyShapeError:
            pass
        return BoundingBox2d()

    def transform_inplace(self, m: Matrix44) -> None:
        self.path.transform_inplace(m)


class FilledPathsRecord(DataRecord):
    def __init__(self, paths: Sequence[NumpyPath2d]) -> None:
        super().__init__()
        self.paths: Sequence[NumpyPath2d] = paths

    def bbox(self) -> BoundingBox2d:
        bbox = BoundingBox2d()
        for path in self.paths:
            if len(path):
                bbox.extend(path.extents())
        return bbox

    def transform_inplace(self, m: Matrix44) -> None:
        for path in self.paths:
            path.transform_inplace(m)


class ImageRecord(DataRecord):
    def __init__(self, boundary: NumpyPoints2d, image_data: ImageData) -> None:
        super().__init__()
        self.boundary: NumpyPoints2d = boundary
        self.image_data: ImageData = image_data

    def bbox(self) -> BoundingBox2d:
        try:
            return self.boundary.bbox()
        except EmptyShapeError:
            pass
        return BoundingBox2d()

    def transform_inplace(self, m: Matrix44) -> None:
        self.boundary.transform_inplace(m)
        self.image_data.transform @= m


class Recorder(BackendInterface):
    """Records the output of the Frontend class."""

    def __init__(self) -> None:
        self.config = Configuration()
        self.background: Color = "#000000"
        self.records: list[DataRecord] = []
        self.properties: dict[int, BackendProperties] = dict()

    def player(self) -> Player:
        """Returns a :class:`Player` instance with the original recordings! Make a copy
        of this player to protect the original recordings from being modified::

            safe_player = recorder.player().copy()

        """
        player = Player()
        player.config = self.config
        player.background = self.background
        player.records = self.records
        player.properties = self.properties
        player.has_shared_recordings = True
        return player

    def configure(self, config: Configuration) -> None:
        self.config = config

    def set_background(self, color: Color) -> None:
        self.background = color

    def store(self, record: DataRecord, properties: BackendProperties) -> None:
        # exclude top-level entity handle to reduce the variance:
        # color, lineweight, layer, pen
        prop_hash = hash(properties[:4])
        record.property_hash = prop_hash
        record.handle = properties.handle
        self.records.append(record)
        self.properties[prop_hash] = properties

    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        self.store(PointsRecord(NumpyPoints2d((pos,))), properties)

    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        self.store(PointsRecord(NumpyPoints2d((start, end))), properties)

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        def flatten() -> Iterator[Vec2]:
            for s, e in lines:
                yield s
                yield e

        self.store(SolidLinesRecord(NumpyPoints2d(flatten())), properties)

    def draw_path(self, path: NumpyPath2d, properties: BackendProperties) -> None:
        assert isinstance(path, NumpyPath2d)
        self.store(PathRecord(path), properties)

    def draw_filled_polygon(
        self, points: NumpyPoints2d, properties: BackendProperties
    ) -> None:
        assert isinstance(points, NumpyPoints2d)
        self.store(PointsRecord(points), properties)

    def draw_filled_paths(
        self, paths: Iterable[NumpyPath2d], properties: BackendProperties
    ) -> None:
        paths = tuple(paths)
        if len(paths) == 0:
            return

        assert isinstance(paths[0], NumpyPath2d)
        self.store(FilledPathsRecord(paths), properties)

    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        # preserve the boundary in image_data in pixel coordinates
        boundary = copy.deepcopy(image_data.pixel_boundary_path)
        boundary.transform_inplace(image_data.transform)
        self.store(ImageRecord(boundary, image_data), properties)

    def enter_entity(self, entity, properties) -> None:
        pass

    def exit_entity(self, entity) -> None:
        pass

    def clear(self) -> None:
        raise NotImplementedError()

    def finalize(self) -> None:
        pass


class Override(NamedTuple):
    """Represents the override state for a data record.

    Attributes:
        properties: original or modified :class:`BackendProperties`
        is_visible: override visibility e.g. switch layers on/off

    """

    properties: BackendProperties
    is_visible: bool = True


OverrideFunc: TypeAlias = Callable[[BackendProperties], Override]


class Player:
    """Plays the recordings of the :class:`Recorder` backend on another backend."""

    def __init__(self) -> None:
        self.config = Configuration()
        self.background: Color = "#000000"
        self.records: list[DataRecord] = []
        self.properties: dict[int, BackendProperties] = dict()
        self._bbox = BoundingBox2d()
        self.has_shared_recordings: bool = False

    def __copy__(self) -> Self:
        """Returns a copy of the player with non-shared recordings."""
        player = self.__class__()
        # config is a frozen dataclass:
        player.config = self.config
        player.background = self.background
        # recordings are mutable: transform and crop inplace
        player.records = copy.deepcopy(self.records)
        # the properties dict may grow, but entries will never be removed:
        player.properties = self.properties
        player.has_shared_recordings = False
        return player

    copy = __copy__

    def recordings(self) -> Iterator[tuple[DataRecord, BackendProperties]]:
        """Yields all recordings as `(DataRecord, BackendProperties)` tuples."""
        props = self.properties
        for record in self.records:
            properties = BackendProperties(
                *props[record.property_hash][:4], record.handle
            )
            yield record, properties

    def replay(
        self, backend: BackendInterface, override: Optional[OverrideFunc] = None
    ) -> None:
        """Replay the recording on another backend that implements the
        :class:`BackendInterface`. The optional `override` function can be used to
        override the properties and state of data records, it gets the :class:`BackendProperties`
        as input and must return an :class:`Override` instance.
        """

        backend.configure(self.config)
        backend.set_background(self.background)
        for record, properties in self.recordings():
            if override:
                state = override(properties)
                if not state.is_visible:
                    continue
                properties = state.properties
            if isinstance(record, PointsRecord):
                count = len(record.points)
                if count == 0:
                    continue
                if count > 2:
                    backend.draw_filled_polygon(record.points, properties)
                    continue
                vertices = record.points.vertices()
                if len(vertices) == 1:
                    backend.draw_point(vertices[0], properties)
                else:
                    backend.draw_line(vertices[0], vertices[1], properties)
            elif isinstance(record, SolidLinesRecord):
                backend.draw_solid_lines(take2(record.lines.vertices()), properties)
            elif isinstance(record, PathRecord):
                backend.draw_path(record.path, properties)
            elif isinstance(record, FilledPathsRecord):
                backend.draw_filled_paths(record.paths, properties)
            elif isinstance(record, ImageRecord):
                backend.draw_image(record.image_data, properties)
        backend.finalize()

    def transform(self, m: Matrix44) -> None:
        """Transforms the recordings inplace by a transformation matrix `m` of type
        :class:`~ezdxf.math.Matrix44`.
        """
        for record in self.records:
            record.transform_inplace(m)

        if self._bbox.has_data:
            # works for 90-, 180- and 270-degree rotation
            self._bbox = BoundingBox2d(m.fast_2d_transform(self._bbox.rect_vertices()))

    def bbox(self) -> BoundingBox2d:
        """Returns the bounding box of all records as :class:`~ezdxf.math.BoundingBox2d`."""
        if not self._bbox.has_data:
            self.update_bbox()
        return self._bbox

    def update_bbox(self) -> None:
        bbox = BoundingBox2d()
        for record in self.records:
            bbox.extend(record.bbox())
        self._bbox = bbox

    def crop_rect(self, p1: UVec, p2: UVec, distance: float) -> None:
        """Crop recorded shapes inplace by a rectangle defined by two points.

        The argument `distance` defines the approximation precision for paths which have
        to be approximated as polylines for cropping but only paths which are really get
        cropped are approximated, paths that are fully inside the crop box will not be
        approximated.

        Args:
            p1: first corner of the clipping rectangle
            p2: second corner of the clipping rectangle
            distance: maximum distance from the center of the curve to the
                center of the line segment between two approximation points to
                determine if a segment should be subdivided.

        """
        crop_rect = BoundingBox2d([Vec2(p1), Vec2(p2)])
        self.records = crop_records_rect(self.records, crop_rect, distance)
        self._bbox = BoundingBox2d()  # determine new bounding box on demand


def crop_records_rect(
    records: list[DataRecord], crop_rect: BoundingBox2d, distance: float
) -> list[DataRecord]:
    """Crop recorded shapes inplace by a rectangle."""

    def sort_paths(np_paths: Sequence[NumpyPath2d]):
        _inside: list[NumpyPath2d] = []
        _crop: list[NumpyPath2d] = []

        for np_path in np_paths:
            bbox = BoundingBox2d(np_path.extents())
            if not crop_rect.has_intersection(bbox):
                # path is complete outside the cropping rectangle
                pass
            elif crop_rect.inside(bbox.extmin) and crop_rect.inside(bbox.extmax):
                # path is complete inside the cropping rectangle
                _inside.append(np_path)
            else:
                _crop.append(np_path)

        return _crop, _inside

    def crop_paths(
        np_paths: Sequence[NumpyPath2d],
    ) -> list[NumpyPath2d]:
        return list(clipper.clip_filled_paths(np_paths, distance))

    # an undefined crop box crops nothing:
    if not crop_rect.has_data:
        return records
    cropped_records: list[DataRecord] = []
    size = crop_rect.size
    # a crop box size of zero in any dimension crops everything:
    if size.x < 1e-12 or size.y < 1e-12:
        return cropped_records

    clipper = ClippingRect(crop_rect.rect_vertices())
    for record in records:
        record_box = record.bbox()
        if not crop_rect.has_intersection(record_box):
            # record is complete outside the cropping rectangle
            continue
        if crop_rect.inside(record_box.extmin) and crop_rect.inside(record_box.extmax):
            # record is complete inside the cropping rectangle
            cropped_records.append(record)
            continue

        if isinstance(record, FilledPathsRecord):
            paths_to_crop, inside = sort_paths(record.paths)
            cropped_paths = crop_paths(paths_to_crop) + inside
            if cropped_paths:
                record.paths = tuple(cropped_paths)
                cropped_records.append(record)
        elif isinstance(record, PathRecord):
            # could be split into multiple parts
            for p in clipper.clip_paths([record.path], distance):
                path_record = PathRecord(p)
                path_record.property_hash = record.property_hash
                path_record.handle = record.handle
                cropped_records.append(path_record)
        elif isinstance(record, PointsRecord):
            count = len(record.points)
            if count == 1:
                # record is inside the clipping shape!
                cropped_records.append(record)
            elif count == 2:
                s, e = record.points.vertices()
                for segment in clipper.clip_line(s, e):
                    if not segment:
                        continue
                    _record = copy.copy(record)  # shallow copy
                    _record.points = NumpyPoints2d(segment)
                    cropped_records.append(_record)
            else:
                for polygon in clipper.clip_polygon(record.points):
                    if not polygon:
                        continue
                    _record = copy.copy(record)  # shallow copy!
                    _record.points = polygon
                    cropped_records.append(_record)
        elif isinstance(record, SolidLinesRecord):
            points: list[Vec2] = []
            for s, e in take2(record.lines.vertices()):
                for segment in clipper.clip_line(s, e):
                    points.extend(segment)
            record.lines = NumpyPoints2d(points)
            cropped_records.append(record)
        elif isinstance(record, ImageRecord):
            pass
            # TODO: Image cropping not supported
            #   Crop image boundary and apply transparency to cropped
            #   parts of the image? -> Image boundary is now a polygon!
        else:
            raise ValueError("invalid record type")
    return cropped_records



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterator, Iterable, Optional
import ezdxf
from ezdxf.addons.xqt import (
    QtWidgets,
    QtGui,
    QAction,
    QMessageBox,
    QFileDialog,
    Qt,
    QModelIndex,
)
from ezdxf.document import Drawing
from ezdxf.entities import Body
from ezdxf.lldxf.const import DXFStructureError
from .data import AcisData, BinaryAcisData, TextAcisData


APP_NAME = "ACIS Structure Browser"
BROWSER_WIDTH = 1024
BROWSER_HEIGHT = 768
SELECTOR_WIDTH_FACTOR = 0.20
FONT_FAMILY = "monospaced"


def make_font():
    font = QtGui.QFont(FONT_FAMILY)
    font.setStyleHint(QtGui.QFont.Monospace)
    return font


class AcisStructureBrowser(QtWidgets.QMainWindow):
    def __init__(
        self,
        filename: str = "",
        handle: str = "",
    ):
        super().__init__()
        self.doc: Optional[Drawing] = None
        self.acis_entities: list[AcisData] = []
        self.current_acis_entity = AcisData()
        self.entity_selector = self.make_entity_selector()
        self.acis_content_viewer = self.make_content_viewer()
        self.statusbar = QtWidgets.QStatusBar(self)
        self.setup_actions()
        self.setup_menu()

        if filename:
            self.load_dxf(filename)
        else:
            self.setWindowTitle(APP_NAME)

        self.setStatusBar(self.statusbar)
        self.setCentralWidget(self.make_central_widget())
        self.resize(BROWSER_WIDTH, BROWSER_HEIGHT)
        self.connect_slots()
        if handle:
            try:
                int(handle, 16)
            except ValueError:
                msg = f"Given handle is not a hex value: '{handle}'"
                self.statusbar.showMessage(msg)
                print(msg)
            else:
                if not self.goto_handle(handle):
                    msg = f"Handle '{handle}' not found."
                    self.statusbar.showMessage(msg)
                    print(msg)

    def make_entity_selector(self):
        return QtWidgets.QListWidget(self)

    def make_content_viewer(self):
        viewer = QtWidgets.QPlainTextEdit(self)
        viewer.setReadOnly(True)
        viewer.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        return viewer

    def make_central_widget(self):
        container = QtWidgets.QSplitter(Qt.Horizontal)
        container.addWidget(self.entity_selector)
        container.addWidget(self.acis_content_viewer)
        selector_width = int(BROWSER_WIDTH * SELECTOR_WIDTH_FACTOR)
        entity_view_width = BROWSER_WIDTH - selector_width
        container.setSizes([selector_width, entity_view_width])
        container.setCollapsible(0, False)
        container.setCollapsible(1, False)
        return container

    def connect_slots(self):
        self.entity_selector.clicked.connect(self.acis_entity_activated)
        self.entity_selector.activated.connect(self.acis_entity_activated)

    # noinspection PyAttributeOutsideInit
    def setup_actions(self):
        self._open_action = self.make_action(
            "&Open DXF File...", self.open_dxf, shortcut="Ctrl+O"
        )
        self._reload_action = self.make_action(
            "Reload DXF File",
            self.reload_dxf,
            shortcut="Ctrl+R",
        )
        self._export_entity_action = self.make_action(
            "&Export Current Entity View...",
            self.export_entity,
            shortcut="Ctrl+E",
        )
        self._export_raw_data_action = self.make_action(
            "&Export Raw SAT/SAB Data...",
            self.export_raw_entity,
            shortcut="Ctrl+W",
        )
        self._quit_action = self.make_action(
            "&Quit", self.close, shortcut="Ctrl+Q"
        )

    def make_action(
        self,
        name,
        slot,
        *,
        shortcut: str = "",
        tip: str = "",
    ) -> QAction:
        action = QAction(name, self)
        if shortcut:
            action.setShortcut(shortcut)
        if tip:
            action.setToolTip(tip)
        action.triggered.connect(slot)
        return action

    def setup_menu(self):
        menu = self.menuBar()
        file_menu = menu.addMenu("&File")
        file_menu.addAction(self._open_action)
        file_menu.addAction(self._reload_action)
        file_menu.addSeparator()
        file_menu.addAction(self._export_entity_action)
        file_menu.addAction(self._export_raw_data_action)
        file_menu.addSeparator()
        file_menu.addAction(self._quit_action)

    def open_dxf(self):
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            caption="Select DXF file",
            filter="DXF Documents (*.dxf *.DXF)",
        )
        if path:
            self.load_dxf(path)

    def load_dxf(self, path: str):
        try:
            doc = ezdxf.readfile(path)
        except IOError as e:
            QMessageBox.critical(self, "Loading Error", str(e))
            return
        except DXFStructureError as e:
            QMessageBox.critical(
                self,
                "DXF Structure Error",
                f'Invalid DXF file "{path}": {str(e)}',
            )
            return
        entities = list(get_acis_entities(doc))
        if len(entities):
            self.doc = doc
            self.set_acis_entities(entities)
            self.update_title(path)
            self.statusbar.showMessage(self.make_loading_message())
        else:
            msg = f"DXF file '{path}' contains no ACIS data"
            QMessageBox.information(self, "Loading Error", msg)  # type: ignore

    def make_loading_message(self) -> str:
        assert self.doc is not None
        dxfversion = self.doc.dxfversion
        acis_type = "SAB" if dxfversion >= "AC1027" else "SAT"
        return f"Loaded DXF file has version {self.doc.acad_release}/{dxfversion}" \
               f" and contains {acis_type} data"

    def set_acis_entities(self, entities: list[AcisData]):
        self.acis_entities = entities
        self.update_entity_selector(entities)
        self.set_current_acis_entity(entities[0])

    def reload_dxf(self):
        try:
            index = self.acis_entities.index(self.current_acis_entity)
        except IndexError:
            index = -1
        self.load_dxf(self.doc.filename)
        if index > 0:
            self.set_current_acis_entity(self.acis_entities[index])

    def export_entity(self):
        dxf_entity = self.get_current_dxf_entity()
        if dxf_entity is None:
            return
        path, _ = QFileDialog.getSaveFileName(
            self,
            caption="Export Current Entity View",
            dir=f"{dxf_entity.dxftype()}-{dxf_entity.dxf.handle}.txt",
            filter="Text Files (*.txt *.TXT)",
        )
        if path:
            write_data(self.current_acis_entity, path)

    def export_raw_entity(self):
        dxf_entity = self.get_current_dxf_entity()
        if dxf_entity is None:
            return
        filename = f"{dxf_entity.dxftype()}-{dxf_entity.dxf.handle}"
        sab = dxf_entity.has_binary_data
        if sab:
            filter_ = "Standard ACIS Binary Files (*.sab *.SAB)"
            filename += ".sab"
        else:
            filter_ = "Standard ACIS Text Files (*.sat *.SAT)"
            filename += ".sat"

        path, _ = QFileDialog.getSaveFileName(
            self,
            caption="Export ACIS Raw Data",
            dir=filename,
            filter=filter_,
        )
        if path:
            if sab:
                with open(path, "wb") as fp:
                    fp.write(dxf_entity.sab)
            else:
                with open(path, "wt") as fp:
                    fp.write("\n".join(dxf_entity.sat))

    def get_current_dxf_entity(self) -> Optional[Body]:
        current = self.current_acis_entity
        if not current.handle or self.doc is None:
            return None
        return self.doc.entitydb.get(current.handle)  # type: ignore

    def update_title(self, path: str):
        self.setWindowTitle(f"{APP_NAME} - {path}")

    def acis_entity_activated(self, index: QModelIndex):
        if len(self.acis_entities) == 0:
            return
        try:
            self.set_current_acis_entity(self.acis_entities[index.row()])
        except IndexError:
            self.set_current_acis_entity(self.acis_entities[0])

    def set_current_acis_entity(self, entity: AcisData):
        if entity:
            self.current_acis_entity = entity
            self.update_acis_content_viewer(entity)

    def update_acis_content_viewer(self, entity: AcisData):
        viewer = self.acis_content_viewer
        viewer.clear()
        viewer.setPlainText("\n".join(entity.lines))

    def update_entity_selector(self, entities: Iterable[AcisData]):
        viewer = self.entity_selector
        viewer.clear()
        viewer.addItems([e.name for e in entities])

    def goto_handle(self, handle: str) -> bool:
        for entity in self.acis_entities:
            if entity.handle == handle:
                self.set_current_acis_entity(entity)
                return True
        return False


def get_acis_entities(doc: Drawing) -> Iterator[AcisData]:
    for e in doc.entitydb.values():
        if isinstance(e, Body):
            handle = e.dxf.handle
            name = f"<{handle}> {e.dxftype()}"
            if e.has_binary_data:
                yield BinaryAcisData(e.sab, name, handle)
            else:
                yield TextAcisData(e.sat, name, handle)


def write_data(entity: AcisData, path: str):
    try:
        with open(path, "wt") as fp:
            fp.write("\n".join(entity.lines))
    except IOError:
        pass




# Copyright (c) 2010-2022 Manfred Moitzi
# License: MIT License
import pytest
import pickle
import math
from ezdxf.math import Vec3, Vec2, Matrix44, ConstructionEllipse, close_vectors

# Import from 'ezdxf.math._bezier4p' to test Python implementation
from ezdxf.math._bezier4p import Bezier4P
from ezdxf.math._bezier4p import cubic_bezier_arc_parameters
from ezdxf.math._bezier4p import cubic_bezier_from_arc
from ezdxf.math._bezier4p import cubic_bezier_from_ellipse
from ezdxf.acc import USE_C_EXT

curve_classes = [Bezier4P]
arc_params_funcs = [cubic_bezier_arc_parameters]
arc_funcs = [cubic_bezier_from_arc]
ellipse_funcs = [cubic_bezier_from_ellipse]

if USE_C_EXT:
    from ezdxf.acc.bezier4p import Bezier4P as CBezier4P
    from ezdxf.acc.bezier4p import (
        cubic_bezier_arc_parameters as cython_arc_parameters,
    )
    from ezdxf.acc.bezier4p import cubic_bezier_from_arc as cython_arc_func
    from ezdxf.acc.bezier4p import (
        cubic_bezier_from_ellipse as cython_ellipse_func,
    )

    curve_classes.append(CBezier4P)
    arc_params_funcs.append(cython_arc_parameters)
    arc_funcs.append(cython_arc_func)
    ellipse_funcs.append(cython_ellipse_func)

DEFPOINTS2D = Vec2.list([(0.0, 0.0), (3.0, 0.0), (7.0, 10.0), (10.0, 10.0)])
DEFPOINTS3D = Vec3.list(
    [
        (0.0, 0.0, 0.0),
        (10.0, 20.0, 20.0),
        (30.0, 10.0, 25.0),
        (40.0, 10.0, 25.0),
    ]
)


@pytest.fixture(params=curve_classes)
def bezier(request):
    return request.param


@pytest.fixture(params=arc_params_funcs)
def arc_params(request):
    return request.param


@pytest.fixture(params=arc_funcs)
def arc(request):
    return request.param


@pytest.fixture(params=ellipse_funcs)
def ellipse(request):
    return request.param


def test_accepts_2d_points(bezier):
    curve = bezier(DEFPOINTS2D)
    for index, chk in enumerate(Vec2.generate(POINTS2D)):
        assert curve.point(index * 0.1).isclose(chk)


def test_objects_are_immutable(bezier):
    curve = bezier(DEFPOINTS3D)
    with pytest.raises(TypeError):
        curve.control_points[0] = (1, 2, 3)


def test_2d_tangent_computation(bezier):
    dbcurve = bezier(DEFPOINTS2D)
    for index, chk in enumerate(Vec2.generate(TANGENTS2D)):
        assert dbcurve.tangent(index * 0.1).isclose(chk)


def test_approximate(bezier):
    curve = bezier(Vec2.list([(0, 0), (0, 1), (1, 1), (1, 0)]))
    with pytest.raises(ValueError):
        list(curve.approximate(0))
    assert list(curve.approximate(1)) == [(0, 0), (1, 0)]
    assert list(curve.approximate(2)) == [(0, 0), (0.5, 0.75), (1, 0)]


def test_reverse(bezier):
    curve = bezier(DEFPOINTS2D)
    vertices = list(curve.approximate(10))
    rev_curve = curve.reverse()
    rev_vertices = list(rev_curve.approximate(10))
    assert close_vectors(reversed(vertices), rev_vertices)


def test_transform_interface(bezier):
    curve = bezier(DEFPOINTS3D)
    new = curve.transform(Matrix44.translate(1, 2, 3))
    assert new.control_points[0] == Vec3(DEFPOINTS3D[0]) + (1, 2, 3)
    assert new.control_points[0] != curve.control_points[0], "expected a new object"


def test_transform_returns_always_3d_curves(bezier):
    curve = bezier(DEFPOINTS2D)
    new = curve.transform(Matrix44.translate(1, 2, 3))
    assert len(new.control_points[0]) == 3


def test_flattening(bezier):
    curve = bezier(Vec2.list([(0, 0), (1, 1), (2, -1), (3, 0)]))
    assert len(list(curve.flattening(1.0, segments=4))) == 5
    assert len(list(curve.flattening(0.1, segments=4))) == 7


def test_flattening_for_equal_start_and_end_points(bezier):
    curve = bezier(Vec2.list([(0, 0), (1, 0), (1, 1), (0, 0)]))
    assert len(list(curve.flattening(0.1, segments=4))) == 5


def test_flattening_with_large_elevation(bezier):
    elevation = 1.391912e19
    curve = bezier(
        Vec3.list(
            [
                (0, 0, elevation),
                (1, 1, elevation),
                (2, -1, elevation),
                (3, 0, elevation),
            ]
        )
    )
    assert len(list(curve.flattening(0.1, segments=4))) > 3


def test_flattening_for_equal_start_and_end_points_large_elevation(bezier):
    elevation = 1.391912e19
    curve = bezier(
        Vec3.list(
            [
                (0, 0, elevation),
                (1, 0, elevation),
                (1, 1, elevation),
                (0, 0, elevation),
            ]
        )
    )
    assert len(list(curve.flattening(0.1, segments=4))) == 5


@pytest.mark.parametrize(
    "z",
    [
        1e99,
        1e79,
        1e59,
        1e39,
        1e19,
        1e9,
        1e6,
        1000,
        0,
        -1e99,
        -1e79,
        -1e59,
        -1e39,
        -1e19,
        -1e9,
        -1e6,
        -1000,
    ],
)
def test_flattening_big_z_coordinates(bezier, z):
    """Test based on issue #574"""
    cp = Vec3.list([
        (888, 770, z),
        (887, 623, z),
        (901, 478, z),
        (930, 335, z),
    ])
    curve = bezier(cp)
    points = list(curve.flattening(0.01))
    # Don't care about the result, it should just not break!
    assert len(points) > 0


def test_pickle_support(bezier):
    curve = bezier(DEFPOINTS3D)
    pickled_curve = pickle.loads(pickle.dumps(curve))
    for index in range(4):
        assert pickled_curve.control_points[index] == DEFPOINTS3D[index]


def test_cubic_bezier_arc_parameters_computation(arc_params):
    parts = list(arc_params(0, math.tau))
    assert len(parts) == 4

    chk = 4.0 * (math.sqrt(2) - 1.0) / 3.0
    sp, cp1, cp2, ep = parts[0]
    assert sp.isclose((1, 0))
    assert cp1.isclose((1, chk))
    assert cp2.isclose((chk, 1))
    assert ep.isclose((0, 1))

    sp, cp1, cp2, ep = parts[1]
    assert sp.isclose((0, 1))
    assert cp1.isclose((-chk, 1))
    assert cp2.isclose((-1, chk))
    assert ep.isclose((-1, 0))

    sp, cp1, cp2, ep = parts[2]
    assert sp.isclose((-1, 0))
    assert cp1.isclose((-1, -chk))
    assert cp2.isclose((-chk, -1))
    assert ep.isclose((0, -1))

    sp, cp1, cp2, ep = parts[3]
    assert sp.isclose((0, -1))
    assert cp1.isclose((chk, -1))
    assert cp2.isclose((1, -chk))
    assert ep.isclose((1, 0))


def test_from_circular_arc(arc):
    curves = list(arc(end_angle=90))
    assert len(curves) == 1

    bezier4p = curves[0]
    cpoints = bezier4p.control_points
    assert len(cpoints) == 4
    assert cpoints[0].isclose((1, 0, 0))
    assert cpoints[1].isclose((1.0, 0.5522847498307933, 0.0))
    assert cpoints[2].isclose((0.5522847498307935, 1.0, 0.0))
    assert cpoints[3].isclose((0, 1, 0))


def test_from_circular_full_arc(arc):
    curves = list(arc(start_angle=180, end_angle=-180))
    assert len(curves) == 4


def test_bezier_curves_from_simple_elliptic_arc(ellipse):
    ellipse_ = ConstructionEllipse(
        center=(1, 1),
        major_axis=(2, 0),
        ratio=0.5,
        start_param=0,
        end_param=math.pi / 2,
    )
    curves = list(ellipse(ellipse_))
    assert len(curves) == 1

    p1, p2, p3, p4 = curves[0].control_points
    assert p1.isclose((3, 1, 0))
    assert p2.isclose((3.0, 1.5522847498307932, 0))
    assert p3.isclose((2.104569499661587, 2.0, 0))
    assert p4.isclose((1, 2, 0))


def test_bezier_curves_from_complex_elliptic_arc(ellipse):
    ellipse_ = ConstructionEllipse(
        center=(49.64089977339618, 36.43095770602131, 0.0),
        major_axis=(16.69099826506408, 6.96203799241026, 0.0),
        ratio=0.173450304570581,
        start_param=5.427509144462117,
        end_param=7.927025930557775,
    )
    curves = list(ellipse(ellipse_))

    assert curves[0].control_points[0].isclose(ellipse_.start_point)
    assert curves[1].control_points[-1].isclose(ellipse_.end_point)


def test_arc_params_issue_708(arc_params):
    cpts = list(arc_params(-2.498091544796509, -0.6435011087932844))
    assert (
        all(
            a.isclose(b)
            for a, b in zip(
                cpts[0],
                (
                    Vec3(-0.8, -0.6, 0.0),
                    Vec3(-0.6111456180001683, -0.8518058426664423, 0.0),
                    Vec3(-0.3147573033330529, -1.0, 0.0),
                    Vec3(6.123233995736766e-17, -1.0, 0.0),
                ),
            )
        )
        is True
    )

    assert (
        all(
            a.isclose(b)
            for a, b in zip(
                cpts[1],
                (
                    Vec3(6.123233995736766e-17, -1.0, 0.0),
                    Vec3(0.314757303333053, -1.0, 0.0),
                    Vec3(0.6111456180001683, -0.8518058426664423, 0.0),
                    Vec3(0.8, -0.5999999999999999, 0.0),
                ),
            )
        )
        is True
    )


def test_bezier_curves_ellipse_issue_708(ellipse):
    ellipse_ = ConstructionEllipse(
        center=(1.5, 0.375, 0.0),
        major_axis=(0.625, 0.0, 0.0),
        ratio=1.0,
        start_param=-2.498091544796509,
        end_param=-0.6435011087932844,
    )
    curves = list(ellipse(ellipse_))
    assert (
        all(
            a.isclose(b)
            for a, b in zip(
                curves[0].control_points,
                (
                    Vec3(0.9999999999999999, 1.6653345369377348e-16, 0.0),
                    Vec3(1.1180339887498947, -0.15737865166652631, 0.0),
                    Vec3(1.3032766854168418, -0.2499999999999999, 0.0),
                    Vec3(1.4999999999999998, -0.25, 0.0),
                ),
            )
        )
        is True
    )
    assert (
        all(
            a.isclose(b)
            for a, b in zip(
                curves[1].control_points,
                (
                    Vec3(1.4999999999999998, -0.25, 0.0),
                    Vec3(1.696723314583158, -0.25, 0.0),
                    Vec3(1.881966011250105, -0.15737865166652654, 0.0),
                    Vec3(2.0, -5.551115123125783e-17, 0.0),
                ),
            )
        )
        is True
    )


POINTS2D = [
    (0.000, 0.000),
    (0.928, 0.280),
    (1.904, 1.040),
    (2.916, 2.160),
    (3.952, 3.520),
    (5.000, 5.000),
    (6.048, 6.480),
    (7.084, 7.840),
    (8.096, 8.960),
    (9.072, 9.720),
    (10.00, 10.00),
]

TANGENTS2D = [
    (9.0, 0.0),
    (9.5400000000000009, 5.3999999999999995),
    (9.9600000000000009, 9.6000000000000014),
    (10.26, 12.600000000000001),
    (10.440000000000001, 14.4),
    (10.5, 15.0),
    (10.44, 14.399999999999999),
    (10.260000000000002, 12.600000000000001),
    (9.9599999999999973, 9.5999999999999925),
    (9.5399999999999974, 5.399999999999995),
    (9.0, 0.0),
]



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.render.linetypes import LineTypeRenderer


def test_line_type_solid():
    ltr = LineTypeRenderer(dashes=tuple())
    assert ltr.is_solid is True
    assert list(ltr.line_segment((0, 0), (5, 0))) == [((0, 0), (5, 0))]


def test_line_start_is_end():
    ltr = LineTypeRenderer(dashes=tuple())
    assert list(ltr.line_segment((0, 0), (0, 0))) == [((0, 0), (0, 0))]


def test_dashed_line_2():
    ltr = LineTypeRenderer(dashes=(1, 1))
    result = list(ltr.line_segment((0, 0), (2, 0)))
    assert len(result) == 1
    assert result[0] == ((0, 0), (1, 0))


def test_dashed_line_4():
    ltr = LineTypeRenderer(dashes=(1, 1))
    result = list(ltr.line_segment((0, 0), (4, 0)))
    assert len(result) == 2
    assert result[0] == ((0, 0), (1, 0))
    # 1,0 -> 2,0 is a gap
    assert result[1] == ((2, 0), (3, 0))


def test_dash_dot_2x_render_issue():
    # DASHDOTX2 (0, 0) -> (20, 0)
    ltr = LineTypeRenderer(dashes=(2.0, 0.2, 0.1, 0.2))
    result = list(ltr.line_segment((0, 0), (20, 0)))
    last_segment = result[-1]
    # Last segment is very, very short:
    assert last_segment[0].isclose(last_segment[1])


if __name__ == "__main__":
    pytest.main([__file__])



# Created: 12.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
from ezdxf.tools.codepage import toencoding, tocodepage


def test_ansi_1250():
    assert "cp1250" == toencoding("ansi_1250")


def test_default():
    assert "cp1252" == toencoding("xyz")


def test_tocodepage_1252():
    assert "ANSI_1252" == tocodepage("cp1252")


def test_tocodepage_936():
    assert "ANSI_936" == tocodepage("gbk")



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Sequence
import math
from ezdxf.math import Vec3
from .bezier_interpolation import (
    tangents_cubic_bezier_interpolation,
    cubic_bezier_interpolation,
)
from .construct2d import circle_radius_3p

__all__ = [
    "estimate_tangents",
    "estimate_end_tangent_magnitude",
    "create_t_vector",
    "chord_length",
]


def create_t_vector(fit_points: list[Vec3], method: str) -> list[float]:
    if method == "uniform":
        return uniform_t_vector(len(fit_points))
    elif method in ("distance", "chord"):
        return distance_t_vector(fit_points)
    elif method in ("centripetal", "sqrt_chord"):
        return centripetal_t_vector(fit_points)
    elif method == "arc":
        return arc_t_vector(fit_points)
    else:
        raise ValueError("Unknown method: {}".format(method))


def uniform_t_vector(length: int) -> list[float]:
    n = float(length - 1)
    return [t / n for t in range(length)]


def distance_t_vector(fit_points: list[Vec3]) -> list[float]:
    return _normalize_distances(list(linear_distances(fit_points)))


def centripetal_t_vector(fit_points: list[Vec3]) -> list[float]:
    distances = [
        math.sqrt(p1.distance(p2)) for p1, p2 in zip(fit_points, fit_points[1:])
    ]
    return _normalize_distances(distances)


def _normalize_distances(distances: Sequence[float]) -> list[float]:
    total_length = sum(distances)
    if abs(total_length) <= 1e-12:
        return []
    params: list[float] = [0.0]
    s = 0.0
    for d in distances[:-1]:
        s += d
        params.append(s / total_length)
    params.append(1.0)
    return params


def linear_distances(points: Iterable[Vec3]) -> Iterable[float]:
    prev = None
    for p in points:
        if prev is None:
            prev = p
            continue
        yield prev.distance(p)
        prev = p


def chord_length(points: Iterable[Vec3]) -> float:
    return sum(linear_distances(points))


def arc_t_vector(fit_points: list[Vec3]) -> list[float]:
    distances = list(arc_distances(fit_points))
    return _normalize_distances(distances)


def arc_distances(fit_points: list[Vec3]) -> Iterable[float]:
    p = fit_points

    def _radii() -> Iterable[float]:
        for i in range(len(p) - 2):
            try:
                radius = circle_radius_3p(p[i], p[i + 1], p[i + 2])
            except ZeroDivisionError:
                radius = 0.0
            yield radius

    r: list[float] = list(_radii())
    if len(r) == 0:
        return
    r.append(r[-1])  # 2x last radius
    for k in range(0, len(p) - 1):
        distance = (p[k + 1] - p[k]).magnitude
        rk = r[k]
        if math.isclose(rk, 0):
            yield distance
        else:
            yield math.asin(distance / 2.0 / rk) * 2.0 * rk


def estimate_tangents(
    points: list[Vec3], method: str = "5-points", normalize=True
) -> list[Vec3]:
    """Estimate tangents for curve defined by given fit points.
    Calculated tangents are normalized (unit-vectors).

    Available tangent estimation methods:

        - "3-points": 3 point interpolation
        - "5-points": 5 point interpolation
        - "bezier": tangents from an interpolated cubic bezier curve
        - "diff": finite difference

    Args:
        points: start-, end- and passing points of curve
        method: tangent estimation method
        normalize: normalize tangents if ``True``

    Returns:
        tangents as list of :class:`Vec3` objects

    """
    method = method.lower()
    if method.startswith("bez"):
        return tangents_cubic_bezier_interpolation(points, normalize=normalize)
    elif method.startswith("3-p"):
        return tangents_3_point_interpolation(points, normalize=normalize)
    elif method.startswith("5-p"):
        return tangents_5_point_interpolation(points, normalize=normalize)
    elif method.startswith("dif"):
        return finite_difference_interpolation(points, normalize=normalize)
    else:
        raise ValueError(f"Unknown method: {method}")


def estimate_end_tangent_magnitude(
    points: list[Vec3], method: str = "chord"
) -> tuple[float, float]:
    """Estimate tangent magnitude of start- and end tangents.

    Available estimation methods:

        - "chord": total chord length, curve approximation by straight segments
        - "arc": total arc length, curve approximation by arcs
        - "bezier-n": total length from cubic bezier curve approximation, n
          segments per section

    Args:
        points: start-, end- and passing points of curve
        method: tangent magnitude estimation method

    """
    if method == "chord":
        total_length = sum(p0.distance(p1) for p0, p1 in zip(points, points[1:]))
        return total_length, total_length
    elif method == "arc":
        total_length = sum(arc_distances(points))
        return total_length, total_length
    elif method.startswith("bezier-"):
        count = int(method[7:])
        s = 0.0
        for curve in cubic_bezier_interpolation(points):
            s += sum(linear_distances(curve.approximate(count)))
        return s, s
    else:
        raise ValueError(f"Unknown tangent magnitude calculation method: {method}")


def tangents_3_point_interpolation(
    fit_points: list[Vec3], method: str = "chord", normalize=True
) -> list[Vec3]:
    """Returns from 3 points interpolated and optional normalized tangent
    vectors.
    """
    q = [Q1 - Q0 for Q0, Q1 in zip(fit_points, fit_points[1:])]
    t = list(create_t_vector(fit_points, method))
    delta_t = [t1 - t0 for t0, t1 in zip(t, t[1:])]
    d = [qk / dtk for qk, dtk in zip(q, delta_t)]
    alpha = [dt0 / (dt0 + dt1) for dt0, dt1 in zip(delta_t, delta_t[1:])]
    tangents: list[Vec3] = [Vec3()]  # placeholder
    tangents.extend(
        [(1.0 - alpha[k]) * d[k] + alpha[k] * d[k + 1] for k in range(len(d) - 1)]
    )
    tangents[0] = 2.0 * d[0] - tangents[1]
    tangents.append(2.0 * d[-1] - tangents[-1])
    if normalize:
        tangents = [v.normalize() for v in tangents]
    return tangents


def tangents_5_point_interpolation(
    fit_points: list[Vec3], normalize=True
) -> list[Vec3]:
    """Returns from 5 points interpolated and optional normalized tangent
    vectors.
    """
    n = len(fit_points)
    q = _delta_q(fit_points)

    alpha = list()
    for k in range(n):
        v1 = (q[k - 1].cross(q[k])).magnitude
        v2 = (q[k + 1].cross(q[k + 2])).magnitude
        alpha.append(v1 / (v1 + v2))

    tangents = []
    for k in range(n):
        vk = (1.0 - alpha[k]) * q[k] + alpha[k] * q[k + 1]
        tangents.append(vk)
    if normalize:
        tangents = [v.normalize() for v in tangents]
    return tangents


def _delta_q(points: list[Vec3]) -> list[Vec3]:
    n = len(points)
    q = [Vec3()]  # placeholder
    q.extend([points[k + 1] - points[k] for k in range(n - 1)])
    q[0] = 2.0 * q[1] - q[2]
    q.append(2.0 * q[n - 1] - q[n - 2])  # q[n]
    q.append(2.0 * q[n] - q[n - 1])  # q[n+1]
    q.append(2.0 * q[0] - q[1])  # q[-1]
    return q


def finite_difference_interpolation(
    fit_points: list[Vec3], normalize=True
) -> list[Vec3]:
    f = 2.0
    p = fit_points

    t = [(p[1] - p[0]) / f]
    for k in range(1, len(fit_points) - 1):
        t.append((p[k] - p[k - 1]) / f + (p[k + 1] - p[k]) / f)
    t.append((p[-1] - p[-2]) / f)
    if normalize:
        t = [v.normalize() for v in t]
    return t


def cardinal_interpolation(fit_points: list[Vec3], tension: float) -> list[Vec3]:
    # https://en.wikipedia.org/wiki/Cubic_Hermite_spline
    def tangent(p0, p1):
        return (p0 - p1).normalize(1.0 - tension)

    t = [tangent(fit_points[0], fit_points[1])]
    for k in range(1, len(fit_points) - 1):
        t.append(tangent(fit_points[k + 1], fit_points[k - 1]))
    t.append(tangent(fit_points[-1], fit_points[-2]))
    return t



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import BoundingBox2d
from ezdxf.addons.hpgl2 import api
from ezdxf.addons.hpgl2.properties import RGB, FillType
from ezdxf.addons.hpgl2.backend import Backend, placement_matrix
from ezdxf.addons.hpgl2.deps import Vec2
from ezdxf.addons.hpgl2.page import Page


def hpgl2(s: bytes):
    return b"%1B" + s


def test_parse_hpgl_commands():
    s = b"%-1BBP;IN;DF;LA1,4,2,6;FT1;PS38812,33987;IP0,0,38812,33987;PU;PA0,0;PUSP0PG;"
    commands = api.hpgl2_commands(s)
    assert len(commands) == 12


def test_skip_all_pcl5_commands():
    s = b"%0Axxxxx%1Ayyyy%-1BBPIN%-1Aescape***%1BDF"
    commands = api.hpgl2_commands(s)
    assert len(commands) == 3
    assert commands[0].name == "BP"
    assert commands[1].name == "IN"
    assert commands[2].name == "DF"


def debug():
    from pathlib import Path

    s = Path(r"C:\Users\mozman\Desktop\Outbox\W2 - 6002.plt").read_bytes()
    commands = api.hpgl2_commands(s)
    assert len(commands) == 0


class MyBackend(Backend):
    def __init__(self):
        self.result = []

    def draw_polyline(self, properties, points) -> None:
        self.result.append(["Polyline", points])

    def draw_paths(self, properties, paths, filled: bool) -> None:
        if filled:
            self.result.append(["FilledPath", paths])
        else:
            self.result.append(["OutlinePath", paths])


def plot(s: bytes):
    commands = api.hpgl2_commands(hpgl2(s))
    backend = MyBackend()
    api.Plotter(backend)
    plotter = api.Plotter(backend)
    interpreter = api.Interpreter(plotter)
    interpreter.run(commands)
    return interpreter


def get_result(plotter):
    return plotter.backend.result


class TestRenderEngine:
    @pytest.mark.parametrize("cmd", [b"PD;PA2000,8000;", b"PU;PA2000,8000;"])
    def test_pen_absolute(self, cmd):
        ip = plot(cmd)
        assert ip.plotter.user_location == Vec2(2000, 8000)

    def test_select_pen(self):
        ip = plot(b"SP7;")
        assert ip.plotter.properties.pen_index == 7

        ip = plot(b"PC3,20,30,40;PW0.18,3;SP3;")
        properties = ip.plotter.properties
        assert properties.pen_index == 3
        assert properties.pen_width == 0.18
        assert properties.pen_color == RGB(20, 30, 40)

    def test_set_pen_width(self):
        ip = plot(b"PW0.18,2;")
        properties = ip.plotter.properties

        assert properties.pen_width == properties.DEFAULT_PEN.width
        pen = properties.get_pen(2)
        assert pen.width == 0.18

    def test_set_current_pen_width(self):
        ip = plot(b"PW0.18;")
        assert ip.plotter.properties.pen_width == 0.18

    def test_set_pen_color(self):
        ip = plot(b"PC2,20,30,40;")
        pen = ip.plotter.properties.get_pen(2)
        assert pen.color == RGB(20, 30, 40)
        assert (
            ip.plotter.properties.pen_color == ip.plotter.properties.DEFAULT_PEN.color
        )

    def test_set_fill_type_hatching(self):
        ip = plot(b"FT3,80,45;")
        props = ip.plotter.properties
        assert props.fill_type == FillType.HATCHING
        assert props.fill_hatch_line_spacing == 80
        assert props.fill_hatch_line_angle == 45

    def test_set_fill_type_shading(self):
        ip = plot(b"FT10,37;")
        props = ip.plotter.properties
        assert props.fill_type == FillType.SHADING
        assert props.fill_shading_density == 37

    def test_pen_relative(self):
        ip = plot(b"PA1000,1000;PR500,-500;")
        assert ip.plotter.user_location == Vec2(1500, 500)

    def test_polyline(self):
        ip = plot(b"PD2000,8000,4000,2000,5000,5000;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        points = command[1]
        assert len(points) == 4
        assert points[0] == Vec2(0, 0)
        assert points[3] == Vec2(5000, 5000)
        assert ip.plotter.user_location == Vec2(5000, 5000)

    def test_cubic_bezier_curve_pen_down(self):
        ip = plot(b"PD;BZ2000,8000,4000,2000,5000,5000;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "OutlinePath"

        paths = command[1]
        path0 = paths[0]
        assert path0.start == Vec2(0, 0)
        assert path0.end == Vec2(5000, 5000)
        assert ip.plotter.user_location == Vec2(5000, 5000)

    def test_cubic_bezier_curve_pen_up(self):
        ip = plot(b"PU;BZ2000,8000,4000,2000,5000,5000;")
        assert len(get_result(ip.plotter)) == 0
        assert ip.plotter.user_location == Vec2(5000, 5000)

    def test_circle(self):
        ip = plot(b"PU;PA2000,8000;PD;CI500;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        assert len(command[1]) == 73  # default chord_angle is 5 deg
        assert ip.plotter.user_location == Vec2(2000, 8000)

    def test_abs_arc(self):
        ip = plot(b"PU100,100;PD;AA200,100,-180;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        assert len(command[1]) == 37  # default chord_angle is 5 deg
        assert ip.plotter.user_location.isclose((300, 100))

    def test_rel_arc(self):
        ip = plot(b"PU100,100;PD;AR100,0,-180;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        assert len(command[1]) == 37  # default chord_angle is 5 deg
        assert ip.plotter.user_location.isclose((300, 100))

    def test_abs_arc_three_points_clockwise(self):
        ip = plot(b"PU100,100;PD;AT200,200,300,100;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        points = command[1]
        assert len(points) == 37  # default chord_angle is 5 deg
        assert points[18].isclose((200, 200))
        assert ip.plotter.user_location.isclose((300, 100))

    def test_abs_arc_three_points_counter_clockwise(self):
        ip = plot(b"PU100,100;PD;AT200,0,300,100;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        points = command[1]
        assert len(points) == 37  # default chord_angle is 5 deg
        assert points[18].isclose((200, 0))
        assert ip.plotter.user_location.isclose((300, 100))

    def test_rel_arc_three_points_clockwise(self):
        ip = plot(b"PU100,100;PD;RT100,100,200,0;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"
        points = command[1]
        assert len(points) == 37  # default chord_angle is 5 deg
        assert points[18].isclose((200, 200))
        assert ip.plotter.user_location.isclose((300, 100))

    def test_polyline_encoded(self):
        ip = plot(b"PE7=U^xGIh;")
        command = get_result(ip.plotter)[0]
        assert command[0] == "Polyline"


class TestTokenizer:
    def parse(self, s: bytes):
        return api.hpgl2_commands(s)

    @pytest.mark.parametrize(
        "s",
        [
            b"%-1BBP;",
            b"%0BBP;",
            b"%1BBP;",
            b"%2BBP;",
            b"%3BBP;",
        ],
    )
    def test_escape(self, s):
        result = self.parse(s)
        assert result[0].name == "BP"

    @pytest.mark.parametrize(
        "s",
        [
            b"IN;PU;PD;",
            b"IN;PU;PD",
            b"INPUPD",
            b"IN PU PD",
            b" INPU PD",
        ],
    )
    def test_short_commands(self, s):
        result = self.parse(hpgl2(s))
        assert len(result) == 3
        assert result[0].name == "IN"
        assert result[1].name == "PU"
        assert result[2].name == "PD"

    @pytest.mark.parametrize(
        "s,i",
        [
            (b"PETEST", 0),
            (b"PUPETEST;PD", 1),
            (b"PUPUPETEST;PD", 2),
        ],
    )
    def test_pe_command(self, s, i):
        result = self.parse(hpgl2(s))
        assert result[i].name == "PE"
        assert result[i].args[0] == b"TEST"

    def test_001(self):
        result = self.parse(hpgl2(b"PA;PA2000,8000"))
        assert result[0].name == "PA"
        assert result[1].name == "PA"

    def test_002(self):
        result = self.parse(hpgl2(b"PUSP0PG"))
        assert result[0].name == "PU"
        assert result[1].name == "SP"
        assert result[1].args[0] == b"0"
        assert result[2].name == "PG"

    def test_filter_non_printable_chars_from_pe_command(self):
        result = self.parse(hpgl2(b"PExxxxx\r\nxxxxx;"))
        assert result[0].args[0] == b"xxxxxxxxxx"

    def test_no_hpgl2_data(self):
        # Ecape sequence to enter HPGL2 mode is missing:  "%1B"
        assert len(self.parse(b"ANYTEXT;IN;BP;")) == 0

    @pytest.mark.parametrize("b", [
        b"LB MY label PA381,0SI0.125,0.25DI0,1;",
        b"LBMY label PA381,0SI0.125,0.25DI0,1;"
    ])
    def test_label_with_terminator(self, b):
        commands = self.parse(hpgl2(b))
        assert len(commands) == 4

    @pytest.mark.parametrize(
        "b",
        [
            b"DT\04LB Tue Dec 18 16:20:05 2001 \04PA381,0SI0.125,0.25DI0,1;",
            b"DT\04,0LB Tue Dec 18 16:20:05 2001 \04PA381,0SI0.125,0.25DI0,1;",
            b"DT\04,1;LB Tue Dec 18 16:20:05 2001 \04PA381,0SI0.125,0.25DI0,1;",
        ],
    )
    def test_label_with_custom_terminator(self, b):
        commands = self.parse(hpgl2(b))
        assert len(commands) == 4

    @pytest.mark.parametrize(
        "b",
        [
            b"DT\04LB MY label\04DTLB My other label ",
            b"DT\04LB MY label\04DT;LB My other label ",
        ],
    )
    def test_reset_custom_terminator(self, b):
        commands = self.parse(hpgl2(b))
        assert len(commands) == 2


class TestPageCoordinates:
    @pytest.fixture(scope="class")
    def page(self):
        page_ = Page(1000, 1000)
        page_.set_ucs(Vec2(500, 500), sx=2, sy=3)
        return page_

    def test_user_to_page_coordinates(self, page):
        assert page.page_point(0, 0).isclose((500, 500))
        assert page.page_point(10, 10).isclose((520, 530))

    def test_user_vector_to_page_vector(self, page):
        assert page.page_vector(0, 0).isclose((0, 0))
        assert page.page_vector(10, 10).isclose((20, 30))


class TestPageAnisotropicScaling:
    def test_isotropic_scaling(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_anisotropic_scaling(-10, 10, -10, 10)
        assert page.user_scaling is True
        assert page.page_point(0, 0).isclose((150, 150))
        assert page.page_point(-10, -10).isclose((100, 100))
        assert page.page_point(10, 10).isclose((200, 200))

    def test_anisotropic_scaling(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_anisotropic_scaling(-10, 10, -20, 20)
        assert page.user_scaling is True
        assert page.page_point(0, 0).isclose((150, 150))
        assert page.page_point(-10, -20).isclose((100, 100))
        assert page.page_point(10, 20).isclose((200, 200))

    def test_reverse_anisotropic_scaling(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        # reverse x and y axis:
        page.set_anisotropic_scaling(10, -10, 20, -20)
        assert page.user_scaling is True
        assert page.page_point(0, 0).isclose((150, 150))
        assert page.page_point(10, 20).isclose((100, 100))
        assert page.page_point(-10, -20).isclose((200, 200))


class TestPageIsotropicScaling:
    def test_isotropic_bottom_window_50(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 20, 0, 10)
        assert page.user_scaling is True
        assert page.page_point(0, 0).isclose((100, 125))
        assert page.page_point(10, 5).isclose((150, 150))
        assert page.page_point(20, 10).isclose((200, 175))

    def test_isotropic_bottom_window_0(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 20, 0, 10, 0, 0)
        assert page.page_point(0, 0).isclose((100, 100))
        assert page.page_point(10, 5).isclose((150, 125))
        assert page.page_point(20, 10).isclose((200, 150))

    def test_isotropic_bottom_window_100(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 20, 0, 10, 1, 1)
        assert page.page_point(0, 0).isclose((100, 150))
        assert page.page_point(10, 5).isclose((150, 175))
        assert page.page_point(20, 10).isclose((200, 200))

    def test_isotropic_left_window_50(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 10, 0, 20)
        assert page.page_point(0, 0).isclose((125, 100))
        assert page.page_point(5, 10).isclose((150, 150))
        assert page.page_point(10, 20).isclose((175, 200))

    def test_isotropic_left_window_0(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 10, 0, 20, 0, 0)
        assert page.page_point(0, 0).isclose((100, 100))
        assert page.page_point(5, 10).isclose((125, 150))
        assert page.page_point(10, 20).isclose((150, 200))

    def test_isotropic_left_window_100(self):
        page = Page(1000, 1000)
        page.set_scaling_points(Vec2(100, 100), Vec2(200, 200))
        page.set_isotropic_scaling(0, 10, 0, 20, 1, 1)
        assert page.page_point(0, 0).isclose((150, 100))
        assert page.page_point(5, 10).isclose((175, 150))
        assert page.page_point(10, 20).isclose((200, 200))


def test_arc_angles():
    from ezdxf.addons.hpgl2.plotter import arc_angles

    angles = list(arc_angles(0, 360, 5))
    assert len(angles) == 73
    assert angles[-1] == pytest.approx(360)

    angles = list(arc_angles(0, -360, 5))
    assert len(angles) == 73
    assert angles[-1] == pytest.approx(-360)


def test_sweeping_angle():
    from ezdxf.addons.hpgl2.plotter import sweeping_angle

    assert sweeping_angle(0, 45, 90) == 90
    assert sweeping_angle(90, 45, 0) == -90
    assert sweeping_angle(330, 0, 30) == 60
    assert sweeping_angle(330, 180, 30) == -300
    assert sweeping_angle(30, 0, 330) == -60
    assert sweeping_angle(30, 180, 330) == 300


class TestPlacementMatrix:
    def test_shift_to_origin_Q1(self):
        bbox = BoundingBox2d([(10, 10), (20, 20)])
        m = placement_matrix(bbox)
        assert m.transform((10, 10)).isclose((0, 0))
        assert m.transform((20, 20)).isclose((10, 10))

    def test_shift_to_origin_Q3(self):
        bbox = BoundingBox2d([(-10, -10), (-20, -20)])
        m = placement_matrix(bbox)
        assert m.transform((-10, -10)).isclose((10, 10))
        assert m.transform((-20, -20)).isclose((0, 0))

    def test_rotate(self):
        # size = 10 x 30
        bbox = BoundingBox2d([(10, 10), (20, 40)])
        m = placement_matrix(bbox, rotation=90)
        assert m.transform((10, 10)).isclose((30, 0))
        assert m.transform((20, 40)).isclose((0, 10))


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.tools.test import load_entities
from ezdxf.sections.table import Table, AppIDTable
from ezdxf.lldxf.tagwriter import TagCollector


@pytest.fixture(scope="module")
def table():
    doc = ezdxf.new()
    return doc.appids


def test_table_entry_dxf_type(table):
    assert table.TABLE_TYPE == "APPID"


def test_ac1009_load_table():
    doc = ezdxf.new("R12")
    entities = list(load_entities(AC1009TABLE, "TABLES"))
    table = AppIDTable()
    table.load(doc, iter(entities[1:-1]))  # without SECTION tags and ENDTAB
    assert len(table) == 10


def test_load_table_with_invalid_table_entry():
    """This LAYERS table has an invalid APPID table entry, which should be
    ignored at the loading stage.
    """
    doc = ezdxf.new("R12")
    entities = list(load_entities(INVALID_TABLE_ENTRY, "TABLES"))
    table = Table()
    table.load(doc, iter(entities[1:-1]))  # without SECTION tags and ENDTAB
    assert len(table) == 0


def test_ac1009_write(table):
    collector = TagCollector(dxfversion="AC1009")
    table.export_dxf(collector)
    tags = collector.tags
    assert tags[0] == (0, "TABLE")
    assert tags[1] == (2, "APPID")
    # exporting table entries is tested by associated class tests
    assert tags[-1] == (0, "ENDTAB")


def test_ac1024_load_table():
    doc = ezdxf.new("R2010")
    entities = list(load_entities(AC1024TABLE, "TABLES"))
    table = AppIDTable()
    table.load(doc, iter(entities[1:-1]))  # without SECTION tags and ENDTAB
    assert 10 == len(table)


def test_ac1024_write(table):
    collector = TagCollector(dxfversion="R2004")
    table.export_dxf(collector)
    tags = collector.tags
    assert tags[0] == (0, "TABLE")
    assert tags[1] == (2, "APPID")
    # exporting table entries is tested by associated class tests
    assert tags[-1] == (0, "ENDTAB")


def test_get_table_entry(table):
    entry = table.get("ACAD")
    assert "ACAD" == entry.dxf.name


def test_entry_names_are_case_insensitive(table):
    entry = table.get("acad")
    assert "ACAD" == entry.dxf.name


def test_duplicate_entry(table):
    new_entry = table.duplicate_entry("ACAD", "ACAD2018")
    assert new_entry.dxf.name == "ACAD2018"

    entry2 = table.get("ACAD2018")
    assert new_entry.dxf.handle == entry2.dxf.handle

    new_entry2 = table.duplicate_entry("ACAD2018", "ACAD2019")
    new_entry.dxf.flags = 71
    new_entry2.dxf.flags = 17
    # really different entities
    assert new_entry.dxf.flags == 71
    assert new_entry2.dxf.flags == 17


def test_create_vport_table():
    doc = ezdxf.new()
    assert len(doc.viewports) == 1
    # standard viewport exists
    assert "*Active" in doc.viewports

    # create a multi-viewport configuration
    # create two entries with same name
    vp1 = doc.viewports.new("V1")
    vp2 = doc.viewports.new("V1")
    assert len(doc.viewports) == 3

    # get multi-viewport configuration as list
    conf = doc.viewports.get_config("V1")
    assert len(conf) == 2

    # check handles
    vports = [vp1, vp2]
    assert conf[0] in vports
    assert conf[1] in vports

    assert "Test" not in doc.viewports
    with pytest.raises(ezdxf.DXFTableEntryError):
        _ = doc.viewports.get_config("test")

    # delete: ignore not existing configurations
    with pytest.raises(ezdxf.DXFTableEntryError):
        doc.viewports.delete_config("test")

    # delete multi config
    doc.viewports.delete_config("V1")
    assert len(doc.viewports) == 1


AC1009TABLE = """0
SECTION
2
TABLES
  0
TABLE
  2
APPID
 70
    10
  0
APPID
  2
ACAD
 70
     0
  0
APPID
  2
ACADANNOPO
 70
     0
  0
APPID
  2
ACADANNOTATIVE
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMJAG
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMTALN
 70
     0
  0
APPID
  2
ACAD_MLEADERVER
 70
     0
  0
APPID
  2
ACAECLAYERSTANDARD
 70
     0
  0
APPID
  2
ACAD_EXEMPT_FROM_CAD_STANDARDS
 70
     0
  0
APPID
  2
ACAD_DSTYLE_DIMBREAK
 70
     0
  0
APPID
  2
ACAD_PSEXT
 70
     0
  0
ENDTAB
  0
ENDSEC
"""

AC1024TABLE = """  0
SECTION
2
TABLES
0
TABLE
  2
APPID
  5
9
330
0
100
AcDbSymbolTable
 70
    10
  0
APPID
  5
12
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD
 70
     0
  0
APPID
  5
DD
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
AcadAnnoPO
 70
     0
  0
APPID
  5
DE
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
AcadAnnotative
 70
     0
  0
APPID
  5
DF
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_DSTYLE_DIMJAG
 70
     0
  0
APPID
  5
E0
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_DSTYLE_DIMTALN
 70
     0
  0
APPID
  5
107
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_MLEADERVER
 70
     0
  0
APPID
  5
1B5
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
AcAecLayerStandard
 70
     0
  0
APPID
  5
1BA
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_EXEMPT_FROM_CAD_STANDARDS
 70
     0
  0
APPID
  5
237
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_DSTYLE_DIMBREAK
 70
     0
  0
APPID
  5
28E
330
9
100
AcDbSymbolTableRecord
100
AcDbRegAppTableRecord
  2
ACAD_PSEXT
 70
     0
  0
ENDTAB
  0
ENDSEC
"""

INVALID_TABLE_ENTRY = """0
SECTION
2
TABLES
0
TABLE
2
LAYERS
70
10
0
APPID
2
ACAD
70
0
0
ENDTAB
0
ENDSEC
"""



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.tools.text import TextScanner


class TestTextScanner:
    def test_scan_empty_text(self):
        s = TextScanner("")
        assert s.is_empty is True
        assert s.has_data is False
        assert s.peek() == ""  # empty string signals end of text
        assert s.get() == ""

    def test_non_empty_string_is_not_empty(self):
        s = TextScanner("word")
        assert s.is_empty is False
        assert s.has_data is True

    def test_peeking_next_letter(self):
        s = TextScanner("word")
        assert s.peek() == "w"

    def test_peeking_more_letters_ahead(self):
        s = TextScanner("word")
        assert s.peek(1) == "o"
        assert s.peek(2) == "r"

    def test_peeking_beyond_word_boundaries_returns_empty_string(self):
        s = TextScanner("word")
        assert s.peek(4) == ""

    def test_getting_next_letter_forwards_scan_position(self):
        s = TextScanner("word")
        assert s.get() == "w"
        assert s.peek() == "o"

    def test_consume_one_letter(self):
        s = TextScanner("word")
        s.consume()
        assert s.peek() == "o"

    def test_consume_two_letter(self):
        s = TextScanner("word")
        s.consume(2)
        assert s.peek() == "r"

    def test_getting_all_letters_empties_scanner(self):
        s = TextScanner("w")
        s.get()
        assert s.is_empty is True
        assert s.has_data is False

    def test_can_not_consume_zero_chars(self):
        s = TextScanner("w")
        pytest.raises(ValueError, s.consume, 0)

    def test_can_not_consume_negative_chars(self):
        s = TextScanner("w")
        pytest.raises(ValueError, s.consume, -1)

    def test_can_not_peek_in_reverse_direction(self):
        s = TextScanner("w")
        pytest.raises(ValueError, s.peek, -1)

    def test_find_next_char(self):
        assert TextScanner(";").find(";") == 0
        assert TextScanner("word;").find(";") == 4
        assert TextScanner("word;;").find(";") == 4

    def test_find_ignores_escaped_chars(self):
        assert TextScanner("word\\;;").find(";", escape=True) == 6
        assert TextScanner("word\\n;").find(";", escape=True) == 6
        assert TextScanner("word\\;\\;;").find(";", escape=True) == 8

    def test_find_next_backslash(self):
        assert TextScanner("\\").find("\\") == 0
        assert TextScanner("a\\").find("\\", escape=False) == 1
        assert TextScanner("a\\").find("\\", escape=True) == 1
        assert TextScanner("a\\\\").find("\\", escape=True) == -1
        assert TextScanner("a\\\\\\").find("\\", escape=True) == 3

    def test_not_find_next(self):
        assert TextScanner("").find(";") == -1
        assert TextScanner("word").find(";") == -1
        assert TextScanner("\\;").find(";", escape=True) == -1

    def test_substr_at_the_begin(self):
        assert TextScanner("").substr(0) == ""
        assert TextScanner("a").substr(1) == "a"
        assert (
            TextScanner("ab").substr(1) == "a"
        ), "do not return the char at the index itself"
        assert TextScanner("ab").substr(2) == "ab"
        assert TextScanner("ab").substr(3) == "ab"

    def test_substr_from_consumed_string(self):
        s = TextScanner("abcdef")
        s.consume(2)
        assert s.substr(2) == ""
        assert s.substr(3) == "c"
        assert s.substr(4) == "cd"
        assert s.substr(10) == "cdef"

    def test_substr_index_error(self):
        with pytest.raises(IndexError):
            TextScanner("").substr(-1)
        s = TextScanner("abcd")
        s.consume(2)
        with pytest.raises(IndexError):
            s.substr(1)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import random

import ezdxf
from ezdxf.document import Drawing
from ezdxf import xref, units
from ezdxf.math import Vec3


FLAG_NAME = "Flag"


def test_xref_doc_has_required_properties(host_doc, xref_doc):
    assert xref_doc.dxfversion == host_doc.dxfversion, "same DXF version expected"
    flag = host_doc.blocks.get(FLAG_NAME)
    assert flag is not None
    assert xref_doc.units == flag.units, "same units as source block expected"
    assert xref_doc.units == units.MM, "mm as unity expected"
    assert Vec3(xref_doc.header["$INSBASE"]).isclose(
        (2, 1)
    ), "XREF has invalid base point"


def test_detached_dxf_contains_all_block_entities(xref_doc: Drawing):
    msp = xref_doc.modelspace()
    types = [e.dxftype() for e in msp]
    assert types == ["POLYLINE", "CIRCLE", "ATTDEF", "ATTDEF", "ATTDEF"]


def test_xref_doc_has_required_resources(xref_doc: Drawing):
    assert xref_doc.layers.has_entry("FLAG_SYMBOL")
    assert xref_doc.layers.has_entry("FLAG_ATTRIBS")


def test_source_block_is_empty(host_doc: Drawing):
    flag = host_doc.blocks.get(FLAG_NAME)
    assert flag is not None, "source block should exist"
    assert len(flag) == 0, "source block should be empty"


def get_random_point():
    x = random.randint(-100, 100)
    y = random.randint(-100, 100)
    return x, y


FLAG_SYMBOL = [(0, 0), (0, 5), (4, 3), (0, 3)]
SAMPLE_COORDS = [get_random_point() for x in range(50)]


@pytest.fixture(scope="module")
def host_doc() -> Drawing:
    doc = ezdxf.new("R2007", units=units.M)
    doc.layers.add("FLAGS")
    doc.layers.add("FLAG_SYMBOL")
    doc.layers.add("FLAG_ATTRIBS")
    msp = doc.modelspace()

    flag = doc.blocks.new(name=FLAG_NAME)
    flag.units = units.MM
    flag.base_point = (2, 1)

    # Add dxf entities to the block (the flag).
    flag.add_polyline2d(FLAG_SYMBOL, dxfattribs={"layer": "FLAG_SYMBOL"})
    flag.add_circle(
        center=(0, 0), radius=0.4, dxfattribs={"layer": "FLAG_SYMBOL", "color": 1}
    )

    # Create the ATTRIB templates as ATTDEF entities:
    flag.add_attdef(
        tag="NAME",
        insert=(0.5, -0.5),
        height=0.5,
        dxfattribs={"layer": "FLAG_ATTRIBS", "color": 3},
    )
    flag.add_attdef(
        tag="XPOS",
        insert=(0.5, -1.0),
        height=0.25,
        dxfattribs={"layer": "FLAG_ATTRIBS", "color": 4},
    )
    flag.add_attdef(
        tag="YPOS",
        insert=(0.5, -1.5),
        height=0.25,
        dxfattribs={"layer": "FLAG_ATTRIBS", "color": 4},
    )

    for number, point in enumerate(SAMPLE_COORDS):
        # Create the value dictionary for the ATTRIB entities, key is the tag
        # name of the ATTDEF entity and the value is the content string of the
        # ATTRIB entity:
        values = {
            "NAME": f"P({number + 1})",
            "XPOS": f"x = {point[0]:.3f}",
            "YPOS": f"y = {point[1]:.3f}",
        }
        random_scale = 0.5 + random.random() * 2.0
        block_ref = msp.add_blockref(
            "FLAG", point, dxfattribs={"layer": "FLAGS", "rotation": -15}
        ).set_scale(random_scale)
        block_ref.add_auto_attribs(values)
    doc.set_modelspace_vport(height=200)
    return doc


@pytest.fixture(scope="module")
def xref_doc(host_doc: Drawing) -> Drawing:
    flag = host_doc.blocks.get(FLAG_NAME)
    assert flag is not None
    return xref.detach(flag, xref_filename="flag_block.dxf")


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations

from ezdxf.entities import XRecord
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.types import DXFTag

__all__ = ["RoundtripXRecord"]

SECTION_MARKER_CODE = 102
NOT_FOUND = -1

class RoundtripXRecord:
    """Helper class for ACAD Roundtrip Data.

    The data is stored in an XRECORD, in sections separated by tags
    (102, "ACAD_SECTION_NAME").

    Example for inverted clipping path of SPATIAL_FILTER objects:

    ...
    100
    AcDbXrecord
    280
    1
    102
    ACAD_INVERTEDCLIP_ROUNDTRIP
    10
    399.725563048036
    20
    233.417786599994
    30
    0.0
    ...
    102
    ACAD_INVERTEDCLIP_ROUNDTRIP_COMPARE
    10
    399.725563048036
    20
    233.417786599994
    ...

    """

    def __init__(self, xrecord: XRecord | None = None) -> None:
        if xrecord is None:
            xrecord = XRecord()
        self.xrecord = xrecord

    def has_section(self, key: str) -> bool:
        """Returns True if an entry section for key is present."""
        for code, value in self.xrecord.tags:
            if code == SECTION_MARKER_CODE and value == key:
                return True
        return False

    def set_section(self, key: str, tags: Tags) -> None:
        """Set content of section `key` to `tags`. Replaces the content of an existing section."""
        xrecord_tags = self.xrecord.tags
        start, end = find_section(xrecord_tags, key)
        if start == NOT_FOUND:
            xrecord_tags.append(DXFTag(SECTION_MARKER_CODE, key))
            xrecord_tags.extend(tags)
        else:
            xrecord_tags[start + 1 : end] = tags

    def get_section(self, key: str) -> Tags:
        """Returns the content of section `key`."""
        xrecord_tags = self.xrecord.tags
        start, end = find_section(xrecord_tags, key)
        if start != NOT_FOUND:
            return xrecord_tags[start + 1 : end]
        return Tags()

    def discard(self, key: str) -> None:
        """Removes section `key`, section `key` doesn't have to exist."""
        xrecord_tags = self.xrecord.tags
        start, end = find_section(xrecord_tags, key)
        if start != NOT_FOUND:
            del xrecord_tags[start:end]


def find_section(tags: Tags, key: str) -> tuple[int, int]:
    """Returns the start- and end index of section `key`.

    Returns (-1, -1) if the section does not exist.
    """
    start = NOT_FOUND
    for index, tag in enumerate(tags):
        if tag.code != 102:
            continue
        if tag.value == key:
            start = index
        elif start != NOT_FOUND:
            return start, index
    if start != NOT_FOUND:
        return start, len(tags)
    return NOT_FOUND, NOT_FOUND



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import ezdxf


@pytest.fixture(scope="module", params=["R12", "R2000"])
def doc(request):
    return ezdxf.new(request.param)


def test_new_doc_extents(doc):
    extmin = doc.header["$EXTMIN"]
    extmax = doc.header["$EXTMAX"]
    assert extmin == (1e20, 1e20, 1e20)
    assert extmax == (-1e20, -1e20, -1e20)


def test_new_doc_limits(doc):
    limmin = doc.header["$LIMMIN"]
    limmax = doc.header["$LIMMAX"]
    assert limmin == (0, 0)
    assert limmax == (420, 297)


def test_default_modelspace_extents(doc):
    msp = doc.modelspace()
    extmin = msp.dxf.extmin
    extmax = msp.dxf.extmax
    assert extmin == (1e20, 1e20, 1e20)
    assert extmax == (-1e20, -1e20, -1e20)


def test_default_modelspace_limits(doc):
    msp = doc.modelspace()
    limmin = msp.dxf.limmin
    limmax = msp.dxf.limmax
    assert limmin == (0, 0)
    assert limmax == (420, 297)


def test_default_layout1_extents(doc):
    layout1 = doc.layout("Layout1")
    extmin = layout1.dxf.extmin
    extmax = layout1.dxf.extmax
    assert extmin == (1e20, 1e20, 1e20)
    assert extmax == (-1e20, -1e20, -1e20)


def test_default_layout1_limits(doc):
    layout1 = doc.layout("Layout1")
    limmin = layout1.dxf.limmin
    limmax = layout1.dxf.limmax
    assert limmin == (0, 0)
    assert limmax == (420, 297)

    layout1.reset_paper_limits()
    assert limmin == (0, 0)
    assert limmax == (420, 297)


def test_reset_modelspace_extents(doc):
    extmin = (-100, -100, -100)
    extmax = (100, 100, 100)
    msp = doc.modelspace()
    msp.reset_extents(extmin, extmax)
    assert msp.dxf.extmin == extmin
    assert msp.dxf.extmax == extmax
    doc.update_extents()  # is automatically called by Drawing.write()
    assert doc.header["$EXTMIN"] == extmin
    assert doc.header["$EXTMAX"] == extmax

    # reset to default values:
    msp.reset_extents()
    assert msp.dxf.extmin == (1e20, 1e20, 1e20)
    assert msp.dxf.extmax == (-1e20, -1e20, -1e20)


def test_reset_modelspace_limits(doc):
    limmin = (-10, -10)
    limmax = (300, 50)
    msp = doc.modelspace()
    msp.reset_limits(limmin, limmax)
    assert msp.dxf.limmin == limmin
    assert msp.dxf.limmax == limmax
    doc.update_limits()  # is automatically called by Drawing.write()
    assert doc.header["$LIMMIN"] == limmin
    assert doc.header["$LIMMAX"] == limmax

    # reset to default values:
    msp.reset_limits()
    width = msp.dxf.paper_width
    height = msp.dxf.paper_height
    assert msp.dxf.limmin == (0, 0)
    assert msp.dxf.limmax == (width, height)


def test_default_active_msp_vport_config(doc):
    # A viewport configuration is always a list of one or more VPORT entities:
    vport_config = doc.viewports.get("*ACTIVE")
    assert len(vport_config) == 1
    vport = vport_config[0]

    assert vport.dxf.center == (0, 0)
    assert vport.dxf.height == 1000


def test_default_active_layout1_viewport(doc):
    layout1 = doc.layout("Layout1")
    assert len(layout1.viewports()) == 0, "no default viewport expected"


def test_reset_layout1_active_viewport(doc):
    doc = ezdxf.new()
    layout1 = cast("Paperspace", doc.layout("Layout1"))
    viewport = layout1.reset_main_viewport()
    assert viewport.dxf.center == (202.5, 128.5)
    paper_width = layout1.dxf.paper_width
    paper_height = layout1.dxf.paper_height
    assert viewport.dxf.width == paper_width * 1.1  # AutoCAD default factor
    assert viewport.dxf.height == paper_height * 1.1  # AutoCAD default factor


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest
import math

import ezdxf
from ezdxf.math import Vec3, Matrix44
from ezdxf.entities.dimension import Dimension, linear_measurement
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.render.dim_base import format_text, DXFValueError, apply_dimpost

TEST_CLASS = Dimension
TEST_TYPE = "DIMENSION"

ENTITY_R12 = """0
DIMENSION
5
0
8
0
2
*D0
3
Standard
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
12
0.0
22
0.0
32
0.0
70
0
1

13
0.0
23
0.0
33
0.0
14
0.0
24
0.0
34
0.0
15
0.0
25
0.0
35
0.0
16
0.0
26
0.0
36
0.0
40
1.0
50
0.0
"""

ENTITY_R2000 = """  0
DIMENSION
5
0
330
0
100
AcDbEntity
8
0
100
AcDbDimension
2
*D0
3
Standard
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
70
32
71
5
42
0.0
100
AcDbAlignedDimension
 13
0.0
 23
0.0
 33
0.0
 14
0.0
 24
0.0
 34
0.0
50
0
100
AcDbRotatedDimension
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "defpoint": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.defpoint == (1, 2, 3)
    assert entity.dxf.defpoint.x == 1, "is not Vec3 compatible"
    assert entity.dxf.defpoint.y == 2, "is not Vec3 compatible"
    assert entity.dxf.defpoint.z == 3, "is not Vec3 compatible"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.defpoint == (0, 0, 0)


@pytest.mark.parametrize("txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)])
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    dimension = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    dimension.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    dimension.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_missing_block_geometry_name():
    doc = ezdxf.new()
    dim = Dimension.new(doc=doc)
    assert dim.get_geometry_block() is None


def test_format_text():
    assert format_text(0, dimrnd=0, dimdec=1, dimzin=0) == "0.0"
    assert format_text(0, dimrnd=0, dimdec=1, dimzin=4) == "0"
    assert format_text(0, dimrnd=0, dimdec=1, dimzin=8) == "0"
    assert format_text(100, dimrnd=0.0, dimdec=0, dimzin=8) == "100"
    assert format_text(100, dimrnd=None, dimdec=0, dimzin=8) == "100"
    assert format_text(100, dimrnd=0, dimzin=8) == "100"
    assert format_text(100, dimrnd=None, dimzin=8) == "100"
    assert format_text(1.23, dimrnd=0.5, dimdec=1, dimzin=0) == "1.0"
    assert format_text(1.23, dimrnd=0.5, dimdec=1, dimzin=8) == "1"
    assert format_text(1.23, dimrnd=0.5, dimdec=1, dimzin=12) == "1"
    assert format_text(10.51, dimrnd=0.5, dimdec=2, dimzin=0) == "10.50"
    assert format_text(10.51, dimrnd=0.5, dimdec=2, dimzin=8) == "10.5"
    assert format_text(0.51, dimrnd=0.5, dimdec=2, dimzin=0 == "0.50")
    assert format_text(0.51, dimrnd=0.5, dimdec=2, dimzin=4) == ".50"
    assert format_text(-0.51, dimrnd=0.5, dimdec=2, dimzin=4) == "-.50"
    assert format_text(-0.11, dimrnd=0.1, dimdec=2, dimzin=4) == "-.10"
    assert format_text(-0.51, dimdsep=",") == "-0,51"


def test_apply_dimpost():
    assert apply_dimpost("0°", "<>") == "0°"
    assert apply_dimpost("30°", "x <>") == "x 30°"
    assert apply_dimpost("30°", "<> x") == "30° x"
    assert apply_dimpost("30°", "x <> x") == "x 30° x"
    assert apply_dimpost("-0,51", "! <> m") == "! -0,51 m"
    assert apply_dimpost("-0,51", "><>") == ">-0,51"
    assert apply_dimpost("-0,51", "<><>") == "-0,51<>"  # ignore stupid


def test_apply_dimpost_raises_exception_for_invalid_text_format():
    with pytest.raises(DXFValueError):
        apply_dimpost("x", dimpost="")
    with pytest.raises(DXFValueError):
        apply_dimpost("x", dimpost="<")


def test_linear_measurement_without_ocs():
    measurement = linear_measurement(Vec3(0, 0, 0), Vec3(1, 0, 0))
    assert measurement == 1

    measurement = linear_measurement(
        Vec3(0, 0, 0), Vec3(1, 0, 0), angle=math.radians(45)
    )
    assert math.isclose(measurement, 1.0 / math.sqrt(2.0))

    measurement = linear_measurement(
        Vec3(0, 0, 0), Vec3(1, 0, 0), angle=math.radians(90)
    )
    assert math.isclose(measurement, 0, abs_tol=1e-12)


def test_dimension_transform_interface():
    dim = Dimension()
    dim.dxf.insert = (1, 0, 0)  # OCS point
    dim.dxf.defpoint = (0, 1, 0)  # WCS point
    dim.dxf.angle = 45

    dim.transform(Matrix44.translate(1, 2, 3))
    assert Vec3(2, 2, 3).isclose(dim.dxf.insert)
    assert Vec3(1, 3, 3).isclose(dim.dxf.defpoint)
    assert math.isclose(dim.dxf.angle, 45)

    dim.transform(Matrix44.z_rotate(math.radians(45)))
    assert math.isclose(dim.dxf.angle, 90)


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_audit_invalid_dimstyle(doc):
    msp = doc.modelspace()
    dim = msp.add_linear_dim((5, 2), (0, 0), (10, 0))
    dim.render()

    dim.dimension.dxf.dimstyle = "XYZ"
    auditor = doc.audit()
    assert len(auditor.fixes) == 1
    assert dim.dimension.dxf.dimstyle == "Standard"


def test_audit_invalid_geometry_block(doc):
    msp = doc.modelspace()
    dim = msp.add_linear_dim((5, 2), (0, 0), (10, 0))
    # no geometry block was created without calling dim.render()

    auditor = doc.audit()
    assert len(auditor.fixes) == 1
    assert dim.dimension.is_alive is False, "DIMENSION should be destroyed"


def test_audit_fake_dimensional_constraints(doc):
    msp = doc.modelspace()
    dim = msp.add_linear_dim((5, 2), (0, 0), (10, 0))
    # fake a dimensional constraint
    dimension = dim.dimension
    dimension.dxf.layer = "*ADSK_CONSTRAINTS"
    # no geometry block was created without calling dim.render()

    auditor = doc.audit()
    assert len(auditor.fixes) == 0
    assert dim.dimension.is_alive is True, "dimensional constraint should be preserved"
